import numpy as np
from kipoi.external.concise.seqplotting_deps import add_letter_to_axis, VOCABS, letter_polygons


def center_cmap(cmap, vmax, vmin, center):
    # Centering of the colormap, taken from seaborn._HeatMapper implementation
    import matplotlib as mpl
    vrange = max(vmax - center, center - vmin)
    normlize = mpl.colors.Normalize(center - vrange, center + vrange)
    cmin, cmax = normlize([vmin, vmax])
    cc = np.linspace(cmin, cmax, 256)
    return mpl.colors.ListedColormap(cmap(cc))


def seqlogo_heatmap(letter_heights, heatmap_data, ovlp_var=None, vocab="DNA", ax=None, show_letter_scale=False,
                    cmap=None, cbar=True, cbar_kws=None, cbar_ax=None, limit_region=None, var_box_color="black",
                    show_var_id=True):
    """
    Plot heatmap and seqlogo plot together in one axis.

    # Arguments
        letter_heights: "sequence length" x "vocabulary size" numpy array of seqlogo letter heights
        heatmap_data: "vocabulary size" x "sequence length" numpy array of heatmap values
    Can also contain negative values.
        vocab: str, Vocabulary name. Can be: DNA, RNA, AA, RNAStruct.
        ax: matplotlib axis
    """
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches
    from matplotlib.collections import PatchCollection
    if cmap is None:
        cmap = plt.cm.bwr
    assert heatmap_data.shape[1] == letter_heights.shape[0]
    seq_len = heatmap_data.shape[1]
    vocab_len = len(VOCABS[vocab])
    letter_rescaling = len(VOCABS[vocab])  # This way the heatmap and the letters are the same size on the plot

    # heatmap grid
    grid = np.mgrid[0.5:(seq_len + 0.5):1, -vocab_len:0:1].reshape(2, -1).T
    y_hm_tickpos = (np.arange(-vocab_len, 0, 1) + 0.5)[::-1]  # alphabet position with 0 on top
    y_seqlogo_tickpos = np.array([0, letter_rescaling])  # tuple of where the ticks for the seqlogo should be placed

    if ax is None:
        plt.figure(figsize=(20, 4))
        ax = plt.subplot(1, 1, 1)
    patches = []
    # add a circle
    for pos_tuple in grid:
        rect = mpatches.Rectangle(pos_tuple, 1.0, 1.0, ec="none")
        patches.append(rect)

    # Add colours to the heatmap - flip the alphabet order so that "A" is on top.
    colors = heatmap_data[::-1, :].T.reshape((seq_len * 4))
    # Centre the colours around 0
    cmap_centered = center_cmap(cmap, colors.max(), colors.min(), 0.0)
    collection = PatchCollection(patches, cmap=cmap_centered, alpha=1.0)
    collection.set_array(np.array(colors))

    # add the heatmap to the axis
    hm_ax_collection = ax.add_collection(collection)

    # rescale letters so that they look nice above the heatmap
    letter_heights_rescaled = np.copy(letter_heights)

    letter_height_scaling = (vocab_len / 2)

    if letter_heights_rescaled.min() < 0:
        lh_range = (letter_heights_rescaled.max() - letter_heights_rescaled.min()) / letter_height_scaling
        letter_y_offset = - letter_heights_rescaled.min() / lh_range
        letter_heights_rescaled = letter_heights_rescaled / lh_range
    else:
        letter_y_offset = 0.0
        letter_heights_rescaled /= letter_heights_rescaled.max() / letter_height_scaling

    assert letter_heights.shape[1] == len(VOCABS[vocab])
    x_range = [1, letter_heights.shape[0]]

    for x_pos, heights in enumerate(letter_heights_rescaled):
        letters_and_heights = sorted(zip(heights, list(VOCABS[vocab].keys())))
        y_pos_pos = letter_y_offset
        y_neg_pos = letter_y_offset
        for height, letter in letters_and_heights:
            color = VOCABS[vocab][letter]
            polygons = letter_polygons[letter]
            if height > 0:
                add_letter_to_axis(ax, polygons, color, 0.5 + x_pos, y_pos_pos, height)
                y_pos_pos += height
            else:
                add_letter_to_axis(ax, polygons, color, 0.5 + x_pos, y_neg_pos, height)
                y_neg_pos += height

    ax.set_xlim(x_range[0] - 1, x_range[1] + 1)
    ax.grid(False)
    ax.set_xticks(list(range(*x_range)) + [x_range[-1]])
    ax.set_aspect(aspect='auto', adjustable='box')

    # set the tick labels and make sure only the left axis is displayed
    if show_letter_scale:
        y_ticks = np.concatenate([y_hm_tickpos, y_seqlogo_tickpos])
        yticklabels = list(VOCABS[vocab].keys()) + ["%.2f" % letter_heights.min(), "%.2f" % letter_heights.max()]
        ax.spines['left'].set_bounds(y_seqlogo_tickpos[0], y_seqlogo_tickpos[1])
    else:
        y_ticks = y_hm_tickpos
        yticklabels = list(VOCABS[vocab].keys())
        ax.spines['left'].set_visible(False)

    ax.set_yticks(y_ticks)
    ax.set_yticklabels(yticklabels)
    ax.axes.get_xaxis().set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.autoscale_view()

    if ovlp_var is not None:
        # for every variant draw a rectangle
        for rel_pos, var_id, ref, alt in zip(ovlp_var["varpos_rel"], ovlp_var["id"], ovlp_var["ref"], ovlp_var["alt"]):
            # positions of ref and alt on the heatmap
            # This is for non-flipped alphabet
            # y_ref_lowlim = -vocab_len + list(VOCABS[vocab].keys()).index(ref[0])
            # y_alt_lowlim = -vocab_len + list(VOCABS[vocab].keys()).index(alt[0])
            # This is for the flipped alphabet
            y_ref_lowlim = list(VOCABS[vocab].keys()).index(ref[0]) * (-1) - 1
            y_alt_lowlim = list(VOCABS[vocab].keys()).index(alt[0][0]) * (-1) - 1
            # box drawing
            box_width = len(ref)
            # Deprecated: draw bax around ref and alt.
            # y_lowlim = min(y_ref_lowlim, y_alt_lowlim)
            # box_height = np.abs(y_ref_lowlim - y_alt_lowlim) + 1
            # only draw the box around the alternative allele.
            y_lowlim = y_alt_lowlim
            box_height = 1
            ax.add_patch(
                mpatches.Rectangle((rel_pos + 0.5, y_lowlim), box_width, box_height, fill=False, lw=2,
                                   ec=var_box_color))
            if show_var_id:
                # annotate the box
                ax.annotate(var_id, xy=(rel_pos + box_width + 0.5, y_lowlim + box_height / 2),
                            xytext=(rel_pos + box_width + 0.5 + 2, y_lowlim + box_height / 2),
                            arrowprops=dict(arrowstyle="->", connectionstyle="arc"),
                            bbox=dict(boxstyle="round,pad=.5", fc="0.9", alpha=0.7))

    if limit_region is not None:
        if not isinstance(limit_region, tuple):
            raise Exception("limit_region has to be tuple of (x_min, x_max)")
        ax.set_xlim(limit_region)

    if cbar:
        # Colorbar settings adapted from seaborn._HeatMapper implementation
        import matplotlib as mpl
        cbar_kws = {} if cbar_kws is None else cbar_kws
        cbar_kws.setdefault('ticks', mpl.ticker.MaxNLocator(6))
        cb = ax.figure.colorbar(hm_ax_collection, cbar_ax, ax, **cbar_kws)
        cb.outline.set_linewidth(0)
        # If rasterized is passed to pcolormesh, also rasterize the
        # colorbar to avoid white lines on the PDF rendering
        # if kws.get('rasterized', False):
        cb.solids.set_rasterized(True)
    return ax



from kipoi import readers
from kipoi.postprocessing.variant_effects import ModelInfoExtractor
from kipoi.postprocessing.variant_effects.utils.generic import _get_metadata_name, _get_seq_shape
from kipoi.postprocessing.variant_effects.utils import ReshapeDna
import kipoi
import os
import numpy as np
import logging
from kipoi.utils import cd

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def grad_x_input(input, grad):
    return input * grad


def get_selector(dim, slice_at_dim):
    """
    Place `slice_at_dim` at dimension `dim` for setting or setting items
    """
    if dim >= 0:
        selector = [slice(None) for i in range(dim)] + [slice_at_dim] + [Ellipsis]
    else:
        selector = [Ellipsis] + [slice_at_dim] + [slice(None) for i in range((dim) * (-1) - 1)]
    selector = tuple(selector)
    return selector


class GradPlotter(object):
    """
    Class for plotting gradients. Results can be loaded from a HDF5 file or directly from returns of
    model.input_grad(...)
    """

    def __init__(self, data, model, source="kipoi", grad_preds=None):
        """
        Arguments:
            data: model input data batch 
            model: model name as used for running `model.input_grad(...)`
            source: model source as used for running `model.input_grad(...)`
            grad_preds: return value of `model.input_grad(...)`. Can alternatively already be present in `data`
            argument under the key `preds`. In that case `grad_preds` may be None.
        """
        self.data = data
        if grad_preds is not None:
            self.data['preds'] = grad_preds
        else:
            assert 'preds' in self.data

        # TODO: Instead of copying from kipoi.model should we rather have a get_model_descr
        # TODO-cont: funcion that is also called from get_model
        # Taken from get_model
        source_name = source
        source = kipoi.config.get_source(source)
        md = source.get_model_descr(model)

        if ":" in md.default_dataloader:
            dl_source, dl_path = md.default_dataloader.split(":")
        else:
            dl_source = source_name
            dl_path = md.default_dataloader

        # allow to use relative and absolute paths for referring to the dataloader
        default_dataloader_path = os.path.join("/" + model, dl_path)[1:]
        # This one loads the model!!
        # default_dataloader = kipoi.get_dataloader_factory(default_dataloader_path,
        #                                                  dl_source)

        # TODO: Is there a nicer way of getting ahold of the dataloader description?
        yaml_path = source.pull_dataloader(default_dataloader_path)
        dataloader_dir = os.path.dirname(yaml_path)
        from kipoi.components import DataLoaderDescription
        with cd(dataloader_dir):
            dl = DataLoaderDescription.load(os.path.basename(yaml_path))
            default_dataloader = dl

        try:
            self.mie = ModelInfoExtractor(md, default_dataloader)
        except:
            logger.warn("Model is not enabled for variant effect prediction hence it is unclear whether there is a DNA "
                        "sequence input, so (automatic) seqlogo plots are not available for this model.")
            self.mie = None
        self.md = md
        self.dataloader = default_dataloader

        # how can the correct model input be selected
        self.get_dataset, self.model_input_keylist = self._get_ds_extractor(md.schema.inputs)

    @classmethod
    def from_hdf5(self, results_fname, model, source="kipoi"):
        """
        Arguments:
            results_fname: HDF5 file produced by running `model.input_grad(...)`. The file has to contain
            also the model input!
            model: model name as used for running `model.input_grad(...)`
            source: model source as used for running `model.input_grad(...)`
        """
        data = readers.HDF5Reader.load(results_fname, unflatten=True)
        return self(data, model=model, source=source)

    def _get_ds_extractor(self, model_input_schema):
        from kipoi.components import ArraySchema
        # return data selection function + labels
        if isinstance(model_input_schema, ArraySchema):
            return self._select_ds_ndarray, [model_input_schema.name]
        elif isinstance(model_input_schema, list):
            return self._select_ds_list, [el.name for el in model_input_schema]
        elif isinstance(model_input_schema, dict):
            return self._select_ds_dict, [el for el in model_input_schema]

    def _select_ds_ndarray(self, index, dataset):
        if index is not None:
            assert (index == 0) or (self.model_input_keylist[0] == index)
        return dataset

    def _select_ds_list(self, index, dataset):
        if isinstance(index, int):
            return dataset[index]
        else:
            int_ind = self.model_input_keylist.index(index)
            return dataset[int_ind]

    def _select_ds_dict(self, index, dataset):
        return dataset[index]

    def _verify_model_input(self, model_input):
        if model_input is None:
            if len(self.model_input_keylist) != 1:
                raise Exception("model_input cannot be None for models that have multiple "
                                "inputs: %s" % str(self.model_input_keylist))
            model_input = self.model_input_keylist[0]
        return model_input

    def get_num_samples(self, model_input=None):
        """
        Get number of samples present in the dataset (size of 0th dimension of the model input).
        
        Arguments:
            model_input: Number of samples will be returned for that model input.
        """
        model_input = self._verify_model_input(model_input)
        input = self.get_dataset(model_input, self.data["inputs"])
        return input.shape[0]

    @staticmethod
    def _infer_seq_dim(len, array_shape):
        sel = np.array(array_shape) == len
        if sel.sum() == 1:
            return np.where(sel)[0][0]
        return None

    def _preprocess_values(self, sample, model_input=None, limit_region=None, limit_region_genomic=None,
                           transform_fn=grad_x_input, seq_dim=None, requires_region_info=False,
                           requires_seq_dim=False):
        def raise_missing_metadata():
            raise Exception("limit_region_genomic can only be used with a `model_input` that has an associated"
                            "metadata field with at least the following entries: 'chrom', 'start', 'end' or type"
                            "GenomicRanges.")

        def get_metadata_cse():
            metadata_field = None
            try:
                metadata_field = _get_metadata_name(self.dataloader, model_input)
            except ValueError:
                raise_missing_metadata()
            mf = self.data['metadata'][metadata_field]
            if not all([el in mf for el in ["chr", "start", "end"]]):
                raise_missing_metadata()
            cse = {k: np.squeeze(mf[k]) for k in ["chr", "start", "end"]}
            for k in ["chr", "start", "end"]:
                cse[k] = np.squeeze(mf[k])
                # if only one sample then squeeze does too much, so correct that:
                if len(cse[k].shape) == 0:
                    cse[k] = np.array([cse[k]])
                if len(cse[k].shape) != 1:
                    raise Exception("Invalid metadata format for field ['%s']['%s'] with shape: %s" %
                                    (metadata_field, k, str(mf[k].shape)))
            return cse

        inputs = self.data["inputs"]
        gradients = self.data["preds"]

        model_input = self._verify_model_input(model_input)

        input = self.get_dataset(model_input, inputs)
        gradient = self.get_dataset(model_input, gradients)

        is_onehot_seq = False
        if (self.mie is not None) and (model_input in self.mie.get_mutatable_inputs(only_one_hot=True)):
            is_onehot_seq = True

        if requires_region_info:
            mf = get_metadata_cse()
            mr_chr = mf["chr"][sample]
            mr_start = mf["start"][sample]
            mr_end = mf["end"][sample]
        else:
            mr_chr = None
            mr_start = None
            mr_end = None

        if (limit_region_genomic is not None):
            if not isinstance(limit_region_genomic, tuple):
                raise Exception("`limit_region_genomic` has to be a tuple of (start, end) genomic coordinates!")
            mf = get_metadata_cse()
            mr_chr = mf["chr"][sample]
            mr_start = mf["start"][sample]
            mr_end = mf["end"][sample]
            if any([(el < mr_start) or (el > mr_end) for el in list(limit_region_genomic)]):
                raise Exception("`limit_region_genomic` has to lie within: %s" % str([mr_start, mr_end]))
            limit_region = (limit_region_genomic[0] - mr_start, limit_region_genomic[1] - mr_start,)
            mr_start, mr_end = limit_region_genomic

        elif (limit_region is not None):
            if not isinstance(limit_region, tuple):
                raise Exception("`limit_region` has to be a tuple of (start, end) array indices!")
            if mr_start is not None:
                mr_end = mr_start + limit_region_genomic[1] + 1
                mr_start = mr_start + limit_region_genomic[0]

        if is_onehot_seq:
            # convert to standard layout
            dna_reshaper = ReshapeDna(_get_seq_shape(self.dataloader, model_input))
            input_reshaped = dna_reshaper.to_standard(input)[sample, ...]
            gradient_reshaped = dna_reshaper.to_standard(gradient)[sample, ...]
            seq_dim = 0

            values = transform_fn(input_reshaped, gradient_reshaped)

            if limit_region is not None:
                values = values[limit_region[0]:limit_region[1], :]

            return values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim

        else:
            if requires_seq_dim:
                if seq_dim is None:
                    mf = get_metadata_cse()
                    seq_len = mf["end"][sample] - mf["start"][sample]
                    seq_dim = self._infer_seq_dim(seq_len, input.shape[1:])
                    if seq_dim is None:
                        raise Exception("seq_dim was not defined and could not be inferred for array dimensions %s and"
                                        "sequence length %d. seq_dim is required if `rc_plot` or "
                                        "`limit_region_genomic` or `seq_plotter_obj` on a non-DNAsequence input "
                                        "are used!")

            values = transform_fn(input[sample, ...], gradient[sample, ...])
            if limit_region is not None:
                lr_sel = get_selector(seq_dim, slice(limit_region[0], limit_region[1]))
                values = values.__getitem__(lr_sel)

            return values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim

    def plot(self, sample, model_input=None, ax=None, limit_region=None, limit_region_genomic=None, rc_plot=False,
             transform_fn=grad_x_input, seq_dim=None, additional_plot_fns=None, **heatmap_kwargs):
        """
        Plot grad*input for one sample in the data (batch). If the selected model input is tagged as "DNASequence" and 
        the model variant effect prediction activated (here only necessary for parsing model info), then values 
        returned by `transform_fn` are displayed as Seqlogo plots. The default transform_fn is grad*input.
        If the selected model input is not tagged as "DNASequence" or the model is not activated for variant effect
        prediction then a heatmap will be generated with additional `heatmap_kwargs` if given.
        If a heatmap should be produced then the input can only be the batch sample axis + 2D, otherwise the heatmap
        fails.  
        
        Arguments:
            sample: Sample in the batch (integer)
            model_input: Name of the model input that should be plotted (can be omitted for models with only one
            input)
            ax: axis object to be passed to the plotting functions.
            limit_region: Tuple. Limits the plot to a subset in the sequence dimension (seq_dim)
            limit_region_genomic: Tuple. Like `limit_region`, but genomic coordinates - no chromosome
            rc_plot: Reverse-complement the plot. If model_input is not "DNASequence" then only reverse
            transform_fn: Function fn(input, grad). Default is input*grad
            seq_dim: Dimension of the sequence. Used for reversing the order in `rc_plot`. Can be omitted for
            "DNASequence" model inputs. If not given, but needed it will be attempted to be inferred from array
            dimensions and metadata sequence length. 
            additional_plot_fns: List of functions fn(chrom, start, end, ax) that will be executed after the main
            plotting routine 
        """

        requires_region_info = additional_plot_fns is not None

        values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim = self._preprocess_values(sample,
                                                                                           model_input=model_input,
                                                                                           limit_region=limit_region,
                                                                                           limit_region_genomic=limit_region_genomic,
                                                                                           transform_fn=transform_fn,
                                                                                           seq_dim=seq_dim,
                                                                                           requires_region_info=requires_region_info,
                                                                                           requires_seq_dim=True)

        if ax is None:
            import matplotlib.pyplot as plt
            plt.figure(figsize=(20, 4))
            ax = plt.subplot(1, 1, 1)

        import seaborn as sns
        if is_onehot_seq:
            # Reverse-complement only for the plot itself
            if rc_plot:
                values = values[::-1, ::-1]

            from kipoi.external.concise.seqplotting_deps import seqlogo
            seqlogo(values, ax=ax)
            ax.axes.get_xaxis().set_visible(False)
            sns.despine(trim=True, bottom=True, ax=ax)

        else:
            if rc_plot:
                rc_sel = get_selector(seq_dim, slice(None, None, -1))
                values = values.__getitem__(rc_sel)

            sns.heatmap(values, ax=ax, **heatmap_kwargs)

        if (additional_plot_fns is not None):
            for plot_fn in additional_plot_fns:
                plot_fn(chrom=mr_chr, start=mr_start, end=mr_end, ax=ax)

    def write(self, sample, writer_obj, model_input=None, limit_region=None, limit_region_genomic=None,
              transform_fn=grad_x_input, seq_dim=None):
        """
        Write grad*input for one sample in the data (batch). If the selected model input is tagged as "DNASequence" and 
        the model variant effect prediction activated (here only necessary for parsing model info), then values 
        returned by `transform_fn` are stored as data in the `writer_obj`. The default transform_fn is grad*input.
        If the selected model input is not tagged as "DNASequence" or the model is not activated for variant effect
        prediction then higher-dimensional data will be compressed into a 1D representation by summing over all axes
        except for the one defined in `seq_dim`. 

        Arguments:
            sample: Sample in the batch (integer)
            writer_obj: Is a instance of a subclass of "RegionWriter". "region_write()" will be called once per
            call of this function.
            model_input: Name of the model input that should be written (can be omitted for models with only one
            input)
            limit_region: Tuple. Limits the values to a subset in the sequence dimension (seq_dim)
            limit_region_genomic: Tuple. Like `limit_region`, but genomic coordinates - no chromosome
            transform_fn: Function fn(input, grad). Default is input*grad
            seq_dim: Dimension of the sequence. Used for reversing the order in `rc_plot`. Can be omitted for
            "DNASequence" model inputs. If not given, but needed it will be attempted to be inferred from array
            dimensions and metadata sequence length. 
        """
        values, is_onehot_seq, mr_chr, mr_start, mr_end, seq_dim = self._preprocess_values(sample,
                                                                                           model_input=model_input,
                                                                                           limit_region=limit_region,
                                                                                           limit_region_genomic=limit_region_genomic,
                                                                                           transform_fn=transform_fn,
                                                                                           seq_dim=seq_dim,
                                                                                           requires_region_info=True,
                                                                                           requires_seq_dim=True)

        regions = {"chr": [mr_chr], "start": [mr_start], "end": [mr_end]}

        if is_onehot_seq:
            writer_obj.region_write(regions, values)


        else:
            # now compress them down by summation so that only the seq_dim is left
            values_summed = values
            for dim in range(0, len(values.shape)):
                if dim != seq_dim:
                    values_summed = np.sum(values_summed, axis=dim, keepdims=True)
            values_summed = np.squeeze(values_summed)
            writer_obj.region_write(regions, values_summed)



import copy
import itertools
import logging
import os
import tempfile

import numpy as np
import pandas as pd
from tqdm import tqdm

from kipoi.postprocessing.variant_effects import BedOverlappingRg, SnvCenteredRg, ensure_tabixed_vcf
from kipoi.postprocessing.variant_effects.scores import Logit, get_scoring_fns
from kipoi.postprocessing.variant_effects.utils import select_from_dl_batch, OutputReshaper, default_vcf_id_gen, \
    ModelInfoExtractor, BedWriter, VariantLocalisation
from kipoi.utils import cd
from .snv_predict import SampleCounter, get_genomicranges_line, merge_intervals, get_variants_in_regions_search_vcf, \
    get_variants_in_regions_sequential_vcf, analyse_model_preds, _overlap_vcf_region
from .utils import is_indel_wrapper

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _generate_records_for_all_regions(regions, ref_seq):
    """
    Generate records for all positions covered by the region
    """
    from vcf.model import _Record, _Substitution
    assert isinstance(regions["chr"], list) or isinstance(regions["chr"], np.ndarray)
    assert isinstance(ref_seq, list)
    assert len(regions["chr"]) == len(ref_seq)
    contained_regions = []
    vcf_records = []
    for i, ref_seq_here in enumerate(ref_seq):
        chrom, start, end = regions["chr"][i], regions["start"][i] + 1, regions["end"][i]
        for pos, ref in zip(range(start, end + 1), ref_seq_here.upper()):
            qual = 0
            filt = []
            info = {}
            fmt = None
            sample_indexes = None
            for alt in ["A", "C", "G", "T"]:
                # skip REF/REF variants - they should always be 0 anyways.
                if ref == alt:
                    continue
                ID = ":".join([chrom, str(pos), ref, alt])
                record = _Record(chrom, pos, ID, ref, [_Substitution(alt)], qual, filt, info, fmt,
                                 sample_indexes)
                vcf_records.append(record)
                contained_regions.append(i)
    #
    return vcf_records, contained_regions


def get_variants_for_all_positions(dl_batch, seq_to_meta, ref_sequences, process_lines_preselection=None):
    """
    Function that generates VCF records for all positions in the input sequence(s). This is done by merging the regions.
    When regions are party overlapping then a variant will be tagged with all sequence-fields that participated in the
    merged region, hence not all input regions might be affected by the variant.
    """
    vcf_records = []  # list of vcf records to use
    process_lines = []  # sample id within batch
    process_seq_fields = []  # sequence fields that should be mutated
    #
    meta_to_seq = {v: [k for k in seq_to_meta if seq_to_meta[k] == v] for v in seq_to_meta.values()}
    all_meta_fields = list(set(seq_to_meta.values()))
    #
    if process_lines_preselection is None:
        num_samples_in_batch = len(dl_batch['metadata'][all_meta_fields[0]]["chr"])
        process_lines_preselection = range(num_samples_in_batch)
    #
    # If we should search for the overlapping VCF lines - for every sample collect all region objects
    # under the assumption that all generated sequences have the same number of samples in a batch:
    for line_id in process_lines_preselection:
        # check is there is more than one metadata_field that is used:

        if len(all_meta_fields) > 1:
            # one region per meta_field
            regions_by_meta = {k: get_genomicranges_line(dl_batch['metadata'][k], line_id)
                               for k in all_meta_fields}
            sequence = {k: ref_sequences[k][line_id] for k in ref_sequences}
            # regions_unif: union across all regions. meta_field_unif_r: meta_fields, has the length of regions_unif
            regions_unif, meta_field_unif_r = merge_intervals(regions_by_meta)
            # get list of merged reference sequences as defined by regions_unif.
            merged_seq = merged_intervals_seq(regions_by_meta, sequence, regions_unif, meta_field_unif_r)
        else:
            # Only one meta_field and only one line hence:
            meta_field_unif_r = [all_meta_fields]
            # get respective reference sequence for the DL batch:
            merged_seq = [ref_sequences[all_meta_fields[0]][line_id]]
            # Only one region:
            regions_unif = get_genomicranges_line(dl_batch['metadata'][all_meta_fields[0]], line_id)
        #
        vcf_records_here, process_lines_rel = _generate_records_for_all_regions(regions_unif, merged_seq)
        #
        for rec, sub_line_id in zip(vcf_records_here, process_lines_rel):
            vcf_records.append(rec)
            process_lines.append(line_id)
            metas = []
            for f in meta_field_unif_r[sub_line_id]:
                metas += meta_to_seq[f]
            process_seq_fields.append(metas)
    return vcf_records, process_lines, process_seq_fields


def merged_intervals_seq(ranges_dict, sequence, regions_unif, meta_field_unif_r):
    """
    ranges_dict: Dict of Genomic ranges objects for all the different metadata slots
    sequence: Sequences stored in a dictionary with metadata slots as keys
    regions_unif: A genomic ranges object of the unified ranges
    meta_field_unif_r: Metadata slot names that are associated with a certain region in `regions_unif`
    seq_to_meta: sequence slot keys to their associated metadata slot key
    """

    all_joint_seqs = []
    for reg_i, mf in enumerate(meta_field_unif_r):
        reg_start = regions_unif["start"][reg_i]
        reg_len = regions_unif["end"][reg_i] - reg_start
        joint_seq = np.empty(reg_len, dtype=str)
        joint_seq[:] = ""
        for mf_here in mf:
            rel_start, rel_end = ranges_dict[mf_here]["start"][0] - reg_start, ranges_dict[mf_here]["end"][
                0] - reg_start
            # When generating the merged sequence make sure the overlapping parts of the sequence match up!
            if np.any(joint_seq[rel_start:rel_end] != ""):
                assert all([a == b for a, b in zip(joint_seq[rel_start:rel_end], sequence[mf_here]) if a != ""])
            joint_seq[rel_start:rel_end] = list(sequence[mf_here])
        all_joint_seqs.append("".join(joint_seq.tolist()))
    return all_joint_seqs


def _overlap_bedtools_region(bedtools_obj, regions):
    """
    Overlap a vcf with regions generated by the dataloader
    The region definition is assumed to be 0-based hence it is converted to 1-based for tabix overlaps!
    Returns VCF records
    """
    assert isinstance(regions["chr"], list) or isinstance(regions["chr"], np.ndarray)
    contained_regions = []
    bed_regions = []
    for i in range(len(regions["chr"])):
        chrom, start, end = regions["chr"][i], regions["start"][i] + 1, regions["end"][i]
        region_str = "{0}:{1}-{2}".format(chrom, start, end)
        bf_regions = bedtools_obj.tabix_intervals(region_str)
        for region in bf_regions:
            bed_regions.append(region)
            contained_regions.append(i)
    #
    return bed_regions, contained_regions


def compress_genomicranges_list(input_list):
    """Convert list of genomicranges objects to a single genomicranges object."""
    # TODO - directly use kipoi.metadata.GenomicRanges.collate(input_list) instead of this function
    assert isinstance(input_list, list)
    out_regions = {k: [] for k in ["chr", "start", "end", "strand"]}
    [[out_regions[k].append(v[k][0]) for k in out_regions] for v in input_list]
    return out_regions


def get_overlapping_bed_regions(dl_batch, seq_to_meta, bedtools_obj):
    """
    Function that overlaps metadata ranges with a bed file.
    Regions are not merged prior to overlap with bedtools_obj!

    Arguments:
        dl_batch: batch coming from the dataloader
        seq_to_meta: dictionary that converts model input names to its associated metadata field.
        bedtools_obj: Tabixed bedtools object of the regions that should be investigated.
    """
    bed_regions = []  # list of vcf records to use
    process_lines = []  # sample id within batch
    process_seq_fields = []  # sequence fields that should be mutated
    #
    meta_to_seq = {v: [k for k in seq_to_meta if seq_to_meta[k] == v] for v in seq_to_meta.values()}
    all_meta_fields = list(set(seq_to_meta.values()))
    #
    num_samples_in_batch = len(dl_batch['metadata'][all_meta_fields[0]]["chr"])
    #
    # If we should search for the overlapping VCF lines - for every sample collect all region objects
    # under the assumption that all generated sequences have the same number of samples in a batch:
    for line_id in range(num_samples_in_batch):
        # check is there is more than one metadata_field that is used:
        if len(all_meta_fields) > 1:
            # As opposed to the VCF handling here don't merge intervals... If two metadata fields are overlapping this
            # then they will be tested independently.
            regions_unif_list = [get_genomicranges_line(dl_batch['metadata'][mf], line_id) for mf in all_meta_fields]
            regions_unif = compress_genomicranges_list(regions_unif_list)
            meta_field_unif_r = [[v] for v in all_meta_fields]
        else:
            # Only one meta_field and only one line hence:
            meta_field_unif_r = [all_meta_fields]
            # Only one region:
            regions_unif = get_genomicranges_line(dl_batch['metadata'][all_meta_fields[0]], line_id)
        #
        bed_regions_here, process_lines_rel = _overlap_bedtools_region(bedtools_obj, regions_unif)
        #
        for reg, sub_line_id in zip(bed_regions_here, process_lines_rel):
            bed_regions.append(reg)
            process_lines.append(line_id)
            metas = []
            for f in meta_field_unif_r[sub_line_id]:
                metas += meta_to_seq[f]
            process_seq_fields.append(metas)
    return bed_regions, process_lines, process_seq_fields


def _generate_seq_sets_mutmap_iter(dl_ouput_schema, dl_batch, seq_to_mut, seq_to_meta,
                                   sample_counter, ref_sequences, bedtools_obj=None, vcf_fh=None,
                                   vcf_id_generator_fn=None, vcf_search_regions=False, generate_rc=True,
                                   batch_size=32):
    all_meta_fields = list(set(seq_to_meta.values()))

    num_samples_in_batch = len(dl_batch['metadata'][all_meta_fields[0]]["chr"])

    metadata_ids = sample_counter.get_ids(num_samples_in_batch)

    if "_id" in dl_batch['metadata']:
        metadata_ids = dl_batch['metadata']['id']
        assert num_samples_in_batch == len(metadata_ids)

    # now get the right region from the vcf:
    # list of vcf records to use: vcf_records
    process_ids = None  # id from genomic ranges metadata: process_lines
    # sample id within batch: process_lines
    # sequence fields that should be mutated: process_seq_fields

    query_bed_regions = None
    query_vcf_records = None
    if vcf_fh is not None:
        # This is from the variant effect prediction function - use this to annotate the output in the end...
        if vcf_search_regions:
            query_vcf_records, query_process_lines, query_process_seq_fields = \
                get_variants_in_regions_search_vcf(dl_batch, seq_to_meta, vcf_fh)
        else:
            # vcf_search_regions == False means: rely completely on the variant id
            # so for every sample assert that all metadata ranges ids agree and then find the entry.
            query_vcf_records, query_process_lines, query_process_seq_fields, query_process_ids = \
                get_variants_in_regions_sequential_vcf(dl_batch, seq_to_meta, vcf_fh, vcf_id_generator_fn)
    elif bedtools_obj is not None:
        query_bed_regions, query_process_lines, query_process_seq_fields = \
            get_overlapping_bed_regions(dl_batch, seq_to_meta, bedtools_obj)
    else:
        # No restrictions are given so process all input lines
        query_process_lines = list(range(num_samples_in_batch))

    # Now generate fake variants for all bases on all positions
    # only the "query_process_lines" selected above should be considered!
    vcf_records, process_lines, process_seq_fields = \
        get_variants_for_all_positions(dl_batch, seq_to_meta, ref_sequences, query_process_lines)

    # short-cut if no sequences are left
    if len(process_lines) == 0:
        raise StopIteration

    if process_ids is None:
        process_ids = []
        for line_id in process_lines:
            process_ids.append(metadata_ids[line_id])

    # Generate a batched output
    real_batch_size = batch_size // 2
    if generate_rc:
        real_batch_size = real_batch_size // 2

    if real_batch_size == 0:
        logger.warn("Batch size too small, resetting it to %d." % (2 + int(generate_rc) * 2))
        real_batch_size = 1

    n_batches = len(process_lines) // real_batch_size
    if len(process_lines) % real_batch_size != 0:
        n_batches += 1

    for batch_i in range(n_batches):
        bs, be = batch_i * real_batch_size, min(((batch_i + 1) * real_batch_size), len(process_lines))
        vcf_records_batch = vcf_records[bs:be]
        process_lines_batch = process_lines[bs:be]
        process_seq_fields_batch = process_seq_fields[bs:be]
        process_ids_batch = process_ids[bs:be]

        # Generate 4 copies of the input set. subset datapoints if needed.
        input_set = {}
        seq_dirs = ["fwd"]
        if generate_rc:
            seq_dirs = ["fwd", "rc"]
        for s_dir, allele in itertools.product(seq_dirs, ["ref", "alt"]):
            k = "%s_%s" % (s_dir, allele)
            ds = dl_batch['inputs']
            all_lines = list(range(num_samples_in_batch))
            if process_lines_batch != all_lines:
                # subset or rearrange elements
                ds = select_from_dl_batch(dl_batch['inputs'], process_lines_batch, num_samples_in_batch)
            input_set[k] = copy.deepcopy(ds)

        # input_set matrices now are in the order required for mutation

        all_mut_seq_keys = list(set(itertools.chain.from_iterable(process_seq_fields_batch)))

        # Start from the sequence inputs mentioned in the model.yaml
        for seq_key in all_mut_seq_keys:
            ranges_input_obj = dl_batch['metadata'][seq_to_meta[seq_key]]
            vl = VariantLocalisation()
            vl.append_multi(seq_key, ranges_input_obj, vcf_records_batch,
                            process_lines_batch, process_ids_batch, process_seq_fields_batch)

            # for the individual sequence input key get the correct sequence mutator callable
            dna_mutator = seq_to_mut[seq_key]

            # Actually modify sequences according to annotation
            # two for loops
            for s_dir, allele in itertools.product(seq_dirs, ["ref", "alt"]):
                k = "%s_%s" % (s_dir, allele)
                if isinstance(dl_ouput_schema.inputs, dict):
                    if seq_key not in input_set[k]:
                        raise Exception("Sequence field %s is missing in DataLoader output!" % seq_key)
                    input_set[k][seq_key] = dna_mutator(input_set[k][seq_key], vl, allele, s_dir)
                elif isinstance(dl_ouput_schema.inputs, list):
                    modified_set = []
                    for seq_el, input_schema_el in zip(input_set[k], dl_ouput_schema.inputs):
                        if input_schema_el.name == seq_key:
                            modified_set.append(dna_mutator(seq_el, vl, allele, s_dir))
                        else:
                            modified_set.append(seq_el)
                    input_set[k] = modified_set
                else:
                    input_set[k] = dna_mutator(input_set[k], vl, allele, s_dir)

        #
        # Reformat so that effect prediction function will get its required inputs
        pred_set = {"ref": input_set["fwd_ref"], "alt": input_set["fwd_alt"]}
        if generate_rc:
            pred_set["ref_rc"] = input_set["rc_ref"]
            pred_set["alt_rc"] = input_set["rc_alt"]
        # pred_set["line_id"] = np.array(process_ids_batch).astype(str) # the process_id does not make sense here
        pred_set["line_id"] = np.array([rec.ID for rec in vcf_records_batch])  # use the id of the variant
        pred_set["vcf_records"] = vcf_records_batch
        pred_set["batch_iters"] = list(range(bs, be))
        pred_set["process_line"] = process_lines_batch
        pred_set["query_bed_regions"] = query_bed_regions
        pred_set["query_vcf_records"] = query_vcf_records
        pred_set["query_process_lines"] = query_process_lines
        yield pred_set


def get_ref_seq_from_seq_set(input_set, seq_to_meta, seq_to_str_converter, dl_ouput_schema_inputs):
    meta_to_seq = {v: [k for k in seq_to_meta if seq_to_meta[k] == v] for v in seq_to_meta.values()}
    all_meta_fields = list(set(seq_to_meta.values()))
    str_seqs = {}

    for meta_field in all_meta_fields:
        seq_key = meta_to_seq[meta_field][0]
        in_set = input_set["inputs"]
        if isinstance(dl_ouput_schema_inputs, dict):
            if seq_key not in in_set:
                raise Exception("Sequence field %s is missing in DataLoader output!" % seq_key)
            seq_data = in_set[seq_key]
        elif isinstance(dl_ouput_schema_inputs, list):
            seq_data = in_set[dl_ouput_schema_inputs.index(seq_key)]
        else:
            seq_data = in_set
        strand_info = input_set["metadata"][meta_field]["strand"]
        if isinstance(strand_info, list):
            strand_info = np.array(strand_info)

        if not isinstance(strand_info, np.ndarray):
            logger.warn("Strand in dataloader batch not defined properly for metadata field: %s. "
                        "Assuming all sequences to be in forward direction!" % meta_field)
            is_rc = np.array([False] * input_set["metadata"][meta_field]["start"].shape[0])
        else:
            is_rc = strand_info == "-"

        str_seqs[meta_field] = seq_to_str_converter[seq_key].to_str(seq_data, is_rc=is_rc)
    return str_seqs


# Structure of mutation_map
# For every DL batch entry across all batches:
#   For every metadata class (seq key):
#       Get the DNA sequence as a String
#       Get the genomic region of the metadata
#       Get all variants that overlap the region and reset to relative coordinates
#       Generate the mapping variants versus effects matrix
#       For every scoring method
#           For every model output
#               Extract the effect score


class MutationMapDataMerger(object):
    def __init__(self, seq_to_meta):
        self.predictions = []
        self.pred_sets = []
        self.ref_seqs = []
        self.batch_metadata_list = []
        self.seq_to_meta = seq_to_meta
        self.meta_to_seq = {v: [k for k in seq_to_meta if seq_to_meta[k] == v] for v in seq_to_meta.values()}
        self.mutation_map = None

    def append(self, predictions, pred_set, ref_seq, batch_metadata):
        # append a new prediction to the set.
        # It is necessary that every batch that is appended is complete in terms of dataloader batches!
        self.predictions.append(predictions)
        self.pred_sets.append(pred_set)
        self.ref_seqs.append(ref_seq)
        self.batch_metadata_list.append(batch_metadata)

    def get_merged_data(self):
        # In general it doesn't make much sense to concatenate the batches. The output shall be as one entry per process_line in pred_sets
        # predictions are stored as dictionary (keys = scoring function) of pandas DFs where each column is a model output.
        # pred_sets contain ["line_id", "vcf_records", "process_line", "query_vcf_records", "query_process_lines"]
        #
        # For every line from the DL
        # mutation_map = [process_line across batches: { model input key: { scoring function key: { Model output label: Object}}}]
        if self.mutation_map is not None:
            return self.mutation_map
        mutation_map = []
        for predictions, pred_set, ref_seq, batch_metadata in zip(self.predictions, self.pred_sets,
                                                                  self.ref_seqs, self.batch_metadata_list):
            for process_line in np.unique(pred_set["process_line"]):
                predictions_rowfilt = pred_set["process_line"] == process_line
                query_vcf_recs = None
                if pred_set["query_vcf_records"] is not None:
                    query_vcf_recs = [rec for rec, pl in zip(pred_set["query_vcf_records"],
                                                             pred_set["query_process_lines"]) if pl == process_line]
                pl_mut_maps = {}
                for metadata_key in ref_seq:
                    metadata_subset = get_genomicranges_line(batch_metadata[metadata_key], process_line)
                    subset_keys = ["chr", "start", "end", "strand"]
                    if not (isinstance(metadata_subset["strand"], list) or
                                isinstance(metadata_subset["strand"], np.ndarray)):
                        subset_keys = ["chr", "start", "end"]
                    metadata_subset_dict = {k: metadata_subset[k][0] for k in subset_keys}
                    if "strand" not in metadata_subset_dict:
                        metadata_subset_dict["strand"] = "*"
                    metadata_chrom = metadata_subset["chr"][0]
                    metadata_start = metadata_subset["start"][0]
                    metadata_end = metadata_subset["end"][0]
                    metadata_seqlen = metadata_end - metadata_start
                    ref_seq_here = ref_seq[metadata_key][process_line]
                    query_vcf_recs_rel = {k: [] for k in ["chr", "pos", "ref", "alt", "id", "varpos_rel"]}
                    if query_vcf_recs is not None:
                        for rec in query_vcf_recs:
                            query_vcf_recs_rel["chr"].append(rec.CHROM)
                            query_vcf_recs_rel["pos"].append(rec.POS)
                            query_vcf_recs_rel["ref"].append(rec.REF)
                            query_vcf_recs_rel["alt"].append(rec.ALT)
                            query_vcf_recs_rel["id"].append(rec.ID)
                            # pos is 1-based, metadata_start is 0 based and the relative position is 0-based, hence -1.
                            query_vcf_recs_rel["varpos_rel"].append(int(rec.POS) - metadata_start - 1)
                    # check that number of variants matches the number of bases *3
                    assert predictions_rowfilt.sum() == metadata_seqlen * 3
                    # generate the ids to get the right order of predictions:
                    correct_order_ids = []
                    # REF/REF was not generated, so those can be appended to the results
                    ref_ref_var_ids = []
                    for pos, ref in zip(range(metadata_start + 1, metadata_end + 1), ref_seq_here.upper()):
                        for alt in ["A", "C", "G", "T"]:
                            ID = ":".join([metadata_chrom, str(pos), ref, alt])
                            correct_order_ids.append(ID)
                            if ref == alt:
                                ref_ref_var_ids.append(ID)
                    metadata_mutmap = {}
                    for scoring_fn in predictions:
                        assert predictions[scoring_fn].shape[0] == predictions_rowfilt.shape[0]
                        metadata_mutmap[scoring_fn] = {}
                        # check if output labels are not unique
                        if len(set(predictions[scoring_fn].columns)) != predictions[scoring_fn].shape[1]:
                            predictions[scoring_fn].columns = ["%s_%d" % (d, i) for i, d in
                                                               enumerate(predictions[scoring_fn].columns)]
                            logger.warn("Model output labels are not unique! appending the column number.")
                        for model_output in predictions[scoring_fn]:
                            mutmap_dict = {}
                            preds = predictions[scoring_fn][model_output].loc[predictions_rowfilt]
                            # append predictions for ref/ref variants:
                            ref_ref_preds = pd.Series(0, name=preds.name, index=ref_ref_var_ids)
                            preds = pd.concat([preds, ref_ref_preds])
                            preds = preds.loc[correct_order_ids].values
                            mutmap_dict["mutation_map"] = np.reshape(preds, (metadata_seqlen, 4)).T
                            mutmap_dict["ovlp_var"] = query_vcf_recs_rel
                            mutmap_dict["ref_seq"] = ref_seq_here
                            mutmap_dict["metadata_region"] = metadata_subset_dict
                            metadata_mutmap[scoring_fn][model_output] = mutmap_dict
                    for seq_key in self.meta_to_seq[metadata_key]:
                        pl_mut_maps[seq_key] = metadata_mutmap
                mutation_map.append(pl_mut_maps)
        self.mutation_map = mutation_map
        return mutation_map

    def save_to_file(self, fname):
        # Best is to use a hdf5 file
        from kipoi.postprocessing.variant_effects.utils.generic import write_hdf5
        return write_hdf5(fname, self.get_merged_data())

    def to_plotter(self):
        return MutationMapPlotter(mutation_map=self.get_merged_data())


class MutationMapPlotter(object):
    def __init__(self, mutation_map=None, fname=None):
        if mutation_map is None and fname is None:
            raise Exception("Either mutation_map or fname for a mutation_map file has to be given")
        if mutation_map is not None:
            self.mutation_map = mutation_map
        elif fname is not None:
            from kipoi.postprocessing.variant_effects.utils.generic import read_hdf5
            self.mutation_map = read_hdf5(fname)

    def save_to_file(self, fname):
        # Best is to use a hdf5 file
        from kipoi.postprocessing.variant_effects.utils.generic import write_hdf5
        return write_hdf5(fname, self.mutation_map)

    def plot_mutmap(self, input_entry, model_seq_input, scoring_key, model_output, ax=None, show_letter_scale=False,
                    cmap=None, limit_region=None, limit_region_genomic=None, annotation_vcf=None,
                    annotation_variants=None, ignore_stored_var_annotation=False, rc_plot=False,
                    minimum_letter_height=None, cbar=True, var_box_color="black",
                    show_var_id=True, grad_inp_style_lh=False):
        """
        Generate a mutation map plot
        
        
        input_entry: index of the region generated by the dataloader. (Integer) 
        model_seq_input: Model input sequence key as defined in model.yaml. e.g. `seq`
        scoring_key: Key given for the scoring method. E.g.: `diff`
        model_output: Model output name for which the mutation map should be plotted
        ax: matplotlib axis object
        show_letter_scale: Display letter scale for seqlogo
        cmap: Colourmap for the heatmap
        limit_region: Limit the plotted region: Tuple of (x_start, x_end) where both x_* are integer in [0,sequence_length)
        limit_region_genomic: Like limit_region but a tuple of genomic positions. Values outside the queried
            regions are ignored. Region definition has to 0-based!
        annotation_vcf: VCF used for additional variant annotation in the plot. Only SNVs will be used.
        annotation_variants: dictionary with key: `chr`, `pos`, `id`, `ref`, `alt` and values are lists of strings,
            except for `pos` which is a list of integers of 1-based variant positions.
        ignore_stored_var_annotation: Ignore annotations that have been stored with the mutation map on generation.
        rc_plot: Reverse-complement plotting
        minimum_letter_height: Require a minimum height of the reference base. proportion of maximum letter height.
            e.g. 0.2
        """

        def append_to_ovlp_var(mm_obj, ovlp_var, pos, id, ref, alt):
            seq_len = mm_obj["end"] - mm_obj["start"]
            varpos_rel = pos - mm_obj["start"] - 1  # variant position is 1-based.
            if (varpos_rel > 0) and (varpos_rel < seq_len):
                ovlp_var['varpos_rel'].append(varpos_rel)
                ovlp_var['id'].append(id)
                ovlp_var['ref'].append(ref)
                ovlp_var['alt'].append(alt)

        from kipoi.external.concise.seqplotting_deps import encodeDNA
        from kipoi.postprocessing.variant_effects.utils.plot import seqlogo_heatmap
        import matplotlib.pyplot as plt
        if cmap is None:
            cmap = plt.cm.bwr
        mm_obj = self.mutation_map[input_entry][model_seq_input][scoring_key][model_output]

        if (limit_region_genomic is not None) and isinstance(limit_region_genomic, tuple):
            mr_start = mm_obj["metadata_region"]["start"]
            mr_end = mm_obj["metadata_region"]["end"]
            if any([(el < mr_start) or (el > mr_end) for el in list(limit_region_genomic)]):
                raise Exception("`limit_region_genomic` has to lie within: %s" % str([mr_start, mr_end]))
            limit_region = (limit_region_genomic[0] - mr_start, limit_region_genomic[1] - mr_start,)

        if (limit_region is None) or ((limit_region is not None) and (not isinstance(limit_region, tuple))):
            limit_region = (0, mm_obj['mutation_map'].shape[1])

        # subset to defined subset
        mm_matrix = mm_obj['mutation_map'][:, limit_region[0]:limit_region[1]]
        ref_dna_str = mm_obj["ref_seq"]
        if hasattr(mm_obj["ref_seq"], "decode"):
            ref_dna_str = mm_obj["ref_seq"].decode('UTF-8')
        ref_dna_str = ref_dna_str[limit_region[0]:limit_region[1]]
        metadata_region = copy.deepcopy(mm_obj["metadata_region"])
        metadata_region["end"] = metadata_region["start"] + limit_region[1]
        metadata_region["start"] = metadata_region["start"] + limit_region[0]

        if ignore_stored_var_annotation or mm_obj['ovlp_var'] is None:
            ovlp_var = {"varpos_rel": [], "id": [], "ref": [], "alt": []}
        else:
            ovlp_var = copy.deepcopy(mm_obj['ovlp_var'])
            # correct the pre-computed variant overlap:
            ovlp_var["varpos_rel"] = [el - limit_region[0] for el in ovlp_var["varpos_rel"]]

        if (annotation_variants is not None) and isinstance(annotation_variants, dict):
            if not all([k in annotation_variants for k in ["chr", "pos", "id", "ref", "alt"]]):
                raise Exception('`annotation_variants` has to be a dictionary with keys ["chr", '
                                '"pos", "id", "ref", "alt"]')
            num_entries = list(set([len(v) for v in annotation_variants.values()]))
            if len(num_entries) != 1:
                raise Exception('All entried in `annotation_variants` have to have the same length!')
            for i in range(num_entries[0]):
                if metadata_region["chr"].lstrip("chr") == str(annotation_variants["chr"][i]).lstrip("chr"):
                    append_to_ovlp_var(metadata_region, ovlp_var, annotation_variants["pos"][i],
                                       annotation_variants["id"][i], annotation_variants["ref"][i],
                                       [annotation_variants["alt"][i]])

        if annotation_vcf is not None:
            import cyvcf2
            if not os.path.exists(annotation_vcf):
                raise Exception('`annotation_vcf` path doesn\'t exist!')
            # ensure tabixed
            vcf_path = ensure_tabixed_vcf(annotation_vcf)
            vcf_fh = cyvcf2.VCF(vcf_path, "r")
            # overlap records with region
            reg = {k: [v] for k, v in metadata_region.items()}
            vcf_records, contained_regions = _overlap_vcf_region(vcf_fh, reg)
            for rec in vcf_records:
                append_to_ovlp_var(metadata_region, ovlp_var, rec.POS, str(rec.ID), str(rec.REF),
                                   [str(el) for el in rec.ALT])

        # RC if necessary
        if rc_plot:
            mm_matrix = mm_matrix[::-1, ::-1]
            from .utils.mutators import rc_str
            ref_dna_str = rc_str(ref_dna_str)
            seq_len = mm_matrix.shape[1]
            ovlp_var["varpos_rel"] = [seq_len - el - 1 for el in ovlp_var["varpos_rel"]]
            ovlp_var["ref"] = rc_str(ovlp_var["ref"])
            ovlp_var["alt"] = [rc_str(el) for el in ovlp_var["alt"]]

        # Derive letter heights from the mutation scores.
        onehot_refseq = encodeDNA([str(ref_dna_str).upper()])[0, ...]
        mm_non_na = mm_matrix.copy()
        nans = np.isnan(mm_non_na)
        if np.any(nans):
            logger.warn(
                "There were %d missing values in the mutation map which are reset to 0 for plotting!" % nans.sum())
            mm_non_na[nans] = 0
        if grad_inp_style_lh:
            letter_heights = onehot_refseq * mm_non_na.mean(axis=0)[:, None] * (-1)
            if minimum_letter_height is not None:
                logger.warn("When using grad_inp_style_lh then minimum_letter_height cannot be used.")
        else:
            # mean of the absolute values
            letter_heights = onehot_refseq * np.abs(mm_non_na).mean(axis=0)[:, None]
            if minimum_letter_height is not None:
                if (minimum_letter_height > 1) or (minimum_letter_height < 0):
                    raise Exception("minimum_letter_height has to be a float within [0,1]")
                max_h = letter_heights.max()
                letter_heights = letter_heights * (
                    1 - minimum_letter_height) + onehot_refseq * minimum_letter_height * max_h

        return seqlogo_heatmap(letter_heights, mm_non_na, ovlp_var, vocab="DNA", ax=ax,
                               show_letter_scale=show_letter_scale, cmap=cmap, limit_region=None,
                               cbar=cbar, var_box_color=var_box_color, show_var_id=show_var_id)


class MutationMap(object):
    def __init__(self, model, dataloader, dataloader_args=None, use_dataloader_example_data=False):
        """Generate mutation map

            Prediction of effects of every base at every position of datalaoder input sequences. The regions for which
            the effect scores will be calculated are primarily defined by the dataloader input. 

            # Arguments
                model: A kipoi model handle generated by e.g.: kipoi.get_model()
                dataloader: Dataloader factory generated by e.g.: kipoi.get_dataloader_factory()
                dataloader_args: arguments passed on to the dataloader for sequence generation, arguments
                    mentioned in dataloader.yaml > postprocessing > variant_effects > bed_input will be overwritten
                    by the methods here.
                use_dataloader_example_data: Fill out the missing dataloader arguments with the example values given in 
                    the dataloader.yaml.
            """
        self.model = model
        self.dataloader = dataloader
        self.model_info_extractor = ModelInfoExtractor(model_obj=model, dataloader_obj=dataloader)
        self.exec_files_bed_keys = self.model_info_extractor.get_exec_files_bed_keys()
        if dataloader_args is None:
            self.dataloader_args = {}
        else:
            self.dataloader_args = dataloader_args

        # Copy the missing arguments from the example arguments.
        if use_dataloader_example_data:
            for k in self.dataloader.example_kwargs:
                if k not in self.dataloader_args:
                    self.dataloader_args[k] = self.dataloader.example_kwargs[k]

    def _setup_dataloader_kwargs(self,
                                 vcf_fpath,
                                 bed_fpath,
                                 vcf_to_region,
                                 bed_to_region,
                                 vcf_id_generator_fn):
        """
        Generate the dataloader kwargs. If e.g. the vcf_fpath should only be used for annotation, but not for
        region generation then set vcf_to_region to None.
        """

        import cyvcf2
        from pybedtools import BedTool
        import copy
        # If then where do I have to put my bed file in the command?
        temp_bed3_file = None
        vcf_search_regions = True

        dataloader_args = copy.deepcopy(self.dataloader_args)

        if (vcf_to_region is not None) and (vcf_fpath is not None) and (bed_to_region is not None) and \
                (bed_fpath is not None):
            logger.warn("`vcf_to_region` and `bed_to_region` are both non-None so regions will only be generated "
                        "based on the VCF! Please use either `vcf_to_region` or `bed_to_region` selectively.")

        # If there is a field in the datalaoder arguments for putting the a postprocessing bed file,
        # then generate the bed file.
        if self.exec_files_bed_keys is not None:
            temp_bed3_file = tempfile.mktemp()  # file path of the temp file
            if (vcf_to_region is not None) and (vcf_fpath is not None):
                logger.warn("Using VCF file %s to define the dataloader intervals." % vcf_fpath)
                vcf_search_regions = False
                vcf_fh = cyvcf2.VCF(vcf_fpath, "r")
                with BedWriter(temp_bed3_file) as ofh:
                    for record in vcf_fh:
                        if not is_indel_wrapper(record):
                            region = vcf_to_region(record)
                            id = vcf_id_generator_fn(record)
                            for chrom, start, end in zip(region["chrom"], region["start"], region["end"]):
                                ofh.append_interval(chrom=chrom, start=start, end=end, id=id)
                vcf_fh.close()
            elif (bed_to_region is not None) and (bed_fpath is not None):
                logger.warn("Using bed file %s to define the dataloader intervals." % bed_fpath)
                bedtools_obj = BedTool(bed_fpath)
                with BedWriter(temp_bed3_file) as ofh:
                    for bed_entry in bedtools_obj:
                        # get all the input regions for a given bed entry
                        in_regions = bed_to_region(bed_entry)
                        for i in range(len(in_regions["chrom"])):
                            ofh.append_interval(**{k: v[i] for k, v in in_regions.items()})
            else:
                logger.warn("Keeping bed file regions defined in `dataloader_args`.")
        else:
            if vcf_to_region is not None:
                logger.warn("`vcf_to_region` will be ignored as it was set, but the dataloader does not define "
                            "a bed_input in dataloader.yaml: "
                            "postprocessing > variant_effects > bed_input.")
            if bed_to_region is not None:
                logger.warn("`bed_to_region` will be ignored as it was set, but the dataloader does not define "
                            "a bed_input in dataloader.yaml: "
                            "postprocessing > variant_effects > bed_input.")

        # If there was a field for dumping the region definition bed file, then use it.
        if (self.exec_files_bed_keys is not None) and ((vcf_search_regions is not None) or (bed_to_region is not None)):
            for k in self.exec_files_bed_keys:
                dataloader_args[k] = temp_bed3_file

        return dataloader_args, temp_bed3_file, vcf_search_regions

    def _generate_mutation_map(self,
                               vcf_fpath=None,
                               bed_fpath=None,
                               batch_size=32,
                               num_workers=0,
                               vcf_to_region=None,
                               bed_to_region=None,
                               vcf_id_generator_fn=default_vcf_id_gen,
                               evaluation_function=analyse_model_preds,
                               evaluation_function_kwargs={'diff_types': {'logit': Logit()}}):
        """Generate mutation map
    
            Prediction of effects of every base at every position of datalaoder input sequences. The regions for which
            the effect scores will be calculated are primarily defined by the dataloader input. Alternatively
            dataloader `bed` file inputs can be overwritten either by sequences centered on variants (see arguments
            `vcf_fpath` and `vcf_to_region`) or bed files (see arguments `bed_fpath` and `bed_to_region`). Effect
            scores are returned as `MutationMapPlotter` object which can be saved to an hdf5 file and used for
            plotting. It is important to mention that the order of the scored sequences is the order in which the
            dataloader has produced data input - intersected with `vcf_fpath` or `bed_fpath` if given.
            Only `vcf_fpath` or `bed_fpath` can be set at once, but neither is required.
    
            # Arguments
                vcf_fpath: If defined only genomic regions overlapping the variants in this VCF will be evaluated.
                    Variants defined here will be highlighted in mutation map plots. Only SNVs will be used. If
                    vcf_to_region is defined and the dataloader accepts bed file input then the dataloader bed input
                    file will be overwritten with regions based on variant positions of this VCF.
                bed_fpath: If defined only genomic regions overlapping regions in this bed file will be evaluated. If
                    bed_to_region is defined and the dataloader accepts bed file input then the dataloader bed input
                    file will be overwritten with regions based this (`bed_fpath`) bed file.
                batch_size: Prediction batch size used for calling the data loader. Each batch will be generated in 4
                    mutated states yielding a system RAM consumption of >= 4x batch size.
                num_workers: Number of parallel workers for loading the dataset.
                vcf_to_region: Callable that generates a regions compatible with dataloader/model from a cyvcf2 record
                bed_to_region: Callable that generates a regions compatible with dataloader/model from a bed entry
                vcf_id_generator_fn: Callable that generates a unique ID from a cyvcf2 record, has to be defined if
                    `vcf_fpath` is set.
                evaluation_function: effect evaluation function. Default is `analyse_model_preds`, which will get
                    arguments defined in `evaluation_function_kwargs`
                evaluation_function_kwargs: kwargs passed on to `evaluation_function`.
    
            # Returns
                A `MutationMapPlotter` object containing variant scores.
            """
        import cyvcf2
        from pybedtools import BedTool
        if (bed_fpath is not None) and (vcf_fpath is not None):
            raise Exception("Can't use both `bed_fpath` and `vcf_fpath`.")

        if (vcf_fpath is not None) and (vcf_id_generator_fn is None):
            raise Exception("If `vcf_fpath` is set then also `vcf_id_generator_fn` has to be defined!")

        dataloader_args, temp_bed3_file, vcf_search_regions = self._setup_dataloader_kwargs(vcf_fpath,
                                                                                            bed_fpath,
                                                                                            vcf_to_region,
                                                                                            bed_to_region,
                                                                                            vcf_id_generator_fn)

        model_out_annotation = self.model_info_extractor.get_model_out_annotation()

        out_reshaper = OutputReshaper(self.model.schema.targets)

        seq_to_mut = self.model_info_extractor.seq_input_mutator
        seq_to_meta = self.model_info_extractor.seq_input_metadata
        seq_to_str_converter = self.model_info_extractor.seq_to_str_converter

        # Open vcf again
        vcf_fh = None
        bed_obj = None
        if vcf_fpath is not None:
            vcf_fh = cyvcf2.VCF(vcf_fpath, "r")
        if bed_fpath is not None:
            bed_obj = BedTool(bed_fpath).tabix()

        # pre-process regions
        keys = set()  # what is that?

        sample_counter = SampleCounter()

        mmdm = MutationMapDataMerger(seq_to_meta)

        # TODO - ignore the un-used params?
        it = self.dataloader(**dataloader_args).batch_iter(batch_size=batch_size,
                                                           num_workers=num_workers)
        for i, batch in enumerate(tqdm(it)):

            # get reference sequence for every line in the batch input
            ref_seq_strs = get_ref_seq_from_seq_set(batch, seq_to_meta, seq_to_str_converter,
                                                    self.dataloader.output_schema.inputs)

            eval_kwargs_iter = _generate_seq_sets_mutmap_iter(self.dataloader.output_schema, batch,
                                                              seq_to_mut=seq_to_mut,
                                                              seq_to_meta=seq_to_meta,
                                                              sample_counter=sample_counter,
                                                              ref_sequences=ref_seq_strs,
                                                              bedtools_obj=bed_obj,
                                                              vcf_fh=vcf_fh,
                                                              vcf_id_generator_fn=vcf_id_generator_fn,
                                                              vcf_search_regions=vcf_search_regions,
                                                              generate_rc=self.model_info_extractor.use_seq_only_rc,
                                                              batch_size=batch_size)

            dl_batch_res = []
            # Keep the following metadata entries from the from the lines
            eval_kwargs_noseq = {k: [] for k in ["line_id", "vcf_records", "process_line"]}
            query_vcf_records = None
            query_process_lines = None

            for eval_kwargs in tqdm(eval_kwargs_iter):
                if eval_kwargs is None:
                    # No generated datapoint overlapped any VCF region
                    continue

                if evaluation_function_kwargs is not None:
                    assert isinstance(evaluation_function_kwargs, dict)
                    for k in evaluation_function_kwargs:
                        eval_kwargs[k] = evaluation_function_kwargs[k]

                eval_kwargs["out_annotation_all_outputs"] = model_out_annotation
                res_here = evaluation_function(self.model, output_reshaper=out_reshaper, **eval_kwargs)
                for k in res_here:
                    keys.add(k)
                    res_here[k].index = eval_kwargs["line_id"]

                # save predictions
                dl_batch_res.append(res_here)

                # save metadata for creating mutation maps
                [eval_kwargs_noseq[k].extend(eval_kwargs[k]) for k in eval_kwargs_noseq]
                query_vcf_records = eval_kwargs["query_vcf_records"]
                query_process_lines = eval_kwargs["query_process_lines"]

            # query vcf entries have to appended at the end of the batch again
            eval_kwargs_noseq["query_vcf_records"] = query_vcf_records
            eval_kwargs_noseq["query_process_lines"] = query_process_lines
            # Concatenate results over batches
            dl_batch_res_concatenated = {}
            for k in keys:
                dl_batch_res_concatenated[k] = pd.concat(
                    [inner_batch[k] for inner_batch in dl_batch_res if k in inner_batch])

            # Append results and inputs to mutation map
            mmdm.append(dl_batch_res_concatenated, eval_kwargs_noseq, ref_seq_strs, batch["metadata"])

            if vcf_fh is not None:
                vcf_fh.close()

            try:
                if temp_bed3_file is not None:
                    os.unlink(temp_bed3_file)
            except:
                pass

            return mmdm.to_plotter()

    def query_region(self,
                     chrom,
                     start,
                     end,
                     seq_length=None,
                     scores=["logit_ref", "logit_alt", "ref", "alt", "logit", "diff"],
                     score_kwargs=None,
                     **kwargs):
        """Generate mutation map

        Prediction of effects of every base at every position of datalaoder input sequences. The regions for which
        the effect scores will be calculated are primarily defined by the dataloader input. If the dataloader accepts
        `bed` file inputs then this file will be overwritten with regions defined here of length `seq_length` or 
        the model input sequence length. If that is not available all datalaoder-generated regions that overlap the
        region defined here will be investigated. Effect scores are returned as `MutationMapPlotter` object which can
        be saved to  an hdf5 file and used for plotting. It is important to mention that the order of the scored
        sequences is the order in which the dataloader has produced data input - intersected with the region defined
        here.

        # Arguments
            chrom: Chrosome of region of interest. Assembly is defined by the dataload arguments.
            start: Start of region of interest. Assembly is defined by the dataload arguments.
            end: End of region of interest. Assembly is defined by the dataload arguments.
            seq_length: Optional argument of model sequence length to use if model accepts variable input
            sequence length. 
            scores: list of score names to compute. See kipoi.postprocessing.variant_effects.scores
            score_kwargs: optional, list of kwargs that corresponds to the entries in score.

        # Returns
            A `MutationMapPlotter` object containing variant scores.
        """
        from pybedtools import BedTool
        bed_to_region = None
        bed_region = BedTool("\t".join(["chr" + chrom.lstrip("chr"), str(start), str(end)]), from_string=True)
        if (self.exec_files_bed_keys is not None):
            bed_to_region = BedOverlappingRg(self.model_info_extractor, seq_length=seq_length)
        dts = get_scoring_fns(self.model, scores, score_kwargs)
        mmdm = self._generate_mutation_map(bed_fpath=bed_region.fn,
                                           vcf_to_region=None,
                                           bed_to_region=bed_to_region,
                                           evaluation_function_kwargs={'diff_types': dts},
                                           **kwargs)
        return mmdm

    def query_bed(self,
                  bed_fpath,
                  seq_length=None,
                  scores=["logit_ref", "logit_alt", "ref", "alt", "logit", "diff"],
                  score_kwargs=None,
                  **kwargs):
        """Generate mutation map

        Prediction of effects of every base at every position of datalaoder input sequences. The regions for which
        the effect scores will be calculated are primarily defined by the dataloader input. If the dataloader accepts
        `bed` file inputs then this file will be overwritten with regions defined in `bed_fpath` of length
        `seq_length` or the model input sequence length. If that is not available all datalaoder-generated
        regions that overlap the region defined here will be investigated. Effect scores are returned as
        `MutationMapPlotter` object which can be saved to  an hdf5 file and used for plotting. It is important to
        mention that the order of the scored sequences is the  order in which the dataloader has produced data
        input - intersected with `bed_fpath`. 

        # Arguments
            bed_fpath: Only genomic regions overlapping regions in this bed file will be evaluated. If
                the dataloader accepts bed file input then the dataloader bed input file will be overwritten with
                regions based this (`bed_fpath`) bed file. Assembly is defined by the dataload arguments.
            seq_length: Optional argument of model sequence length to use if model accepts variable input
                sequence length. 
            scores: list of score names to compute. See kipoi.postprocessing.variant_effects.scores
            score_kwargs: optional, list of kwargs that corresponds to the entries in score.

        # Returns
            A `MutationMapPlotter` object containing variant scores.
        """
        bed_to_region = None
        if (self.exec_files_bed_keys is not None):
            bed_to_region = BedOverlappingRg(self.model_info_extractor, seq_length=seq_length)
        dts = get_scoring_fns(self.model, scores, score_kwargs)
        mmdm = self._generate_mutation_map(vcf_fpath=None,
                                           bed_fpath=bed_fpath,
                                           vcf_to_region=None,
                                           bed_to_region=bed_to_region,
                                           vcf_id_generator_fn=default_vcf_id_gen,
                                           evaluation_function_kwargs={'diff_types': dts},
                                           **kwargs)
        return mmdm

    def query_vcf(self,
                  vcf_fpath,
                  seq_length=None,
                  scores=["logit_ref", "logit_alt", "ref", "alt", "logit", "diff"],
                  score_kwargs=None,
                  var_centered_regions=True,
                  **kwargs):
        """Generate mutation map

        Prediction of effects of every base at every position of datalaoder input sequences. The regions for which
        the effect scores will be calculated are primarily defined by the dataloader input. If the dataloader accepts
        `bed` file inputs then this file will be overwritten with regions generaten from the SNVs in `vcf_fpath`in a
        variant-centered fashion. Sequence length is defined by `seq_length` or the model input sequence length.
        If the datalaoder does not have a `bed` file input all datalaoder-generated regions that overlap SNVs
        here will be investigated. Effect scores are returned as `MutationMapPlotter` object which can be saved to 
        an hdf5 file and used for plotting. It is important to mention that the order of the scored sequences is the 
        order in which the dataloader has produced data input - intersected with `vcf_fpath`. 

        # Arguments
            vcf_fpath: Only genomic regions overlapping the variants in this VCF will be evaluated.
                Variants defined here will be highlighted in mutation map plots. Only SNVs will be used. If
                vcf_to_region is defined and the dataloader accepts bed file input then the dataloader bed input
                file will be overwritten with regions based on variant positions of this VCF.
            seq_length: Optional argument of model sequence length to use if model accepts variable input
                sequence length.
            var_centered_regions: Generate variant-centered regions if the model accepts that. If a custom
                `vcf_to_region` should be used then this can be set explicitly in the kwargs.
            scores: list of score names to compute. See kipoi.postprocessing.variant_effects.scores
            score_kwargs: optional, list of kwargs that corresponds to the entries in score.

        # Returns
            A `MutationMapPlotter` object containing variant scores.
        """
        vcf_to_region = None
        if var_centered_regions and (self.exec_files_bed_keys is not None):
            vcf_to_region = SnvCenteredRg(self.model_info_extractor, seq_length=seq_length)
        if "vcf_to_region" in kwargs:
            vcf_to_region = kwargs["vcf_to_region"]
        dts = get_scoring_fns(self.model, scores, score_kwargs)
        mmdm = self._generate_mutation_map(vcf_fpath=vcf_fpath,
                                           bed_fpath=None,
                                           vcf_to_region=vcf_to_region,
                                           bed_to_region=None,
                                           vcf_id_generator_fn=default_vcf_id_gen,
                                           evaluation_function=analyse_model_preds,
                                           evaluation_function_kwargs={'diff_types': dts},
                                           **kwargs)
        return mmdm


def _generate_mutation_map(model,
                           dataloader,
                           vcf_fpath=None,
                           bed_fpath=None,
                           batch_size=32,
                           num_workers=0,
                           dataloader_args=None,
                           vcf_to_region=None,
                           bed_to_region=None,
                           vcf_id_generator_fn=default_vcf_id_gen,
                           evaluation_function=analyse_model_preds,
                           evaluation_function_kwargs={'diff_types': {'logit': Logit()}},
                           use_dataloader_example_data=False,
                           ):
    """Generate mutation map

    Prediction of effects of every base at every position of datalaoder input sequences. The regions for which
    the effect scores will be calculated are primarily defined by the dataloader input. Alternatively
    dataloader `bed` file inputs can be overwritten either by sequences centered on variants (see arguments
    `vcf_fpath` and `vcf_to_region`) or bed files (see arguments `bed_fpath` and `bed_to_region`). Effect
    scores are returned as `MutationMapPlotter` object which can be saved to an hdf5 file and used for
    plotting. It is important to mention that the order of the scored sequences is the order in which the
    dataloader has produced data input - intersected with `vcf_fpath` or `bed_fpath` if given.
    Only `vcf_fpath` or `bed_fpath` can be set at once, but neither is required.

    # Arguments
        model: A kipoi model handle generated by e.g.: kipoi.get_model()
        dataloader: Dataloader factory generated by e.g.: kipoi.get_dataloader_factory()
        vcf_fpath: If defined only genomic regions overlapping the variants in this VCF will be evaluated.
            Variants defined here will be highlighted in mutation map plots. Only SNVs will be used. If
            vcf_to_region is defined and the dataloader accepts bed file input then the dataloader bed input
            file will be overwritten with regions based on variant positions of this VCF.
        bed_fpath: If defined only genomic regions overlapping regions in this bed file will be evaluated. If
            bed_to_region is defined and the dataloader accepts bed file input then the dataloader bed input
            file will be overwritten with regions based this (`bed_fpath`) bed file.
        batch_size: Prediction batch size used for calling the data loader. Each batch will be generated in 4
            mutated states yielding a system RAM consumption of >= 4x batch size.
        num_workers: Number of parallel workers for loading the dataset.
        dataloader_args: arguments passed on to the dataloader for sequence generation, arguments
            mentioned in dataloader.yaml > postprocessing > variant_effects > bed_input will be overwritten
            by the methods here.
        vcf_to_region: Callable that generates a regions compatible with dataloader/model from a cyvcf2 record
        bed_to_region: Callable that generates a regions compatible with dataloader/model from a bed entry
        vcf_id_generator_fn: Callable that generates a unique ID from a cyvcf2 record, has to be defined if
            `vcf_fpath` is set.
        evaluation_function: effect evaluation function. Default is `analyse_model_preds`, which will get
            arguments defined in `evaluation_function_kwargs`
        evaluation_function_kwargs: kwargs passed on to `evaluation_function`.
        use_dataloader_example_data: Fill out the missing dataloader arguments with the example values given in the
            dataloader.yaml.

    # Returns
        A `MutationMapPlotter` object containing variant scores.
    """
    mm = MutationMap(model, dataloader, dataloader_args, use_dataloader_example_data)
    return mm._generate_mutation_map(vcf_fpath=vcf_fpath,
                                     bed_fpath=bed_fpath,
                                     batch_size=batch_size,
                                     num_workers=num_workers,
                                     vcf_to_region=vcf_to_region,
                                     bed_to_region=bed_to_region,
                                     vcf_id_generator_fn=vcf_id_generator_fn,
                                     evaluation_function=evaluation_function,
                                     evaluation_function_kwargs=evaluation_function_kwargs)



import re
from collections import OrderedDict

import numpy as np

from kipoi.external.concise.seqplotting_letters import all_letters


###### Had to copy from concise due to requirements
# -----------------------------------------------------------------------
#
#
# Code adoped from https://github.com/kundajelab/dragonn, (c) 2016 Kundaje Lab


def standardize_polygons_str(data_str):
    """Given a POLYGON string, standardize the coordinates to a 1x1 grid.
    Input : data_str (taken from above)
    Output: tuple of polygon objects
    """
    # find all of the polygons in the letter (for instance an A
    # needs to be constructed from 2 polygons)
    from shapely.wkt import loads as load_wkt
    path_strs = re.findall("\(\(([^\)]+?)\)\)", data_str.strip())

    # convert the data into a numpy array
    polygons_data = []
    for path_str in path_strs:
        data = np.array([
            tuple(map(float, x.split())) for x in path_str.strip().split(",")])
        polygons_data.append(data)

    # standardize the coordinates
    min_coords = np.vstack(data.min(0) for data in polygons_data).min(0)
    max_coords = np.vstack(data.max(0) for data in polygons_data).max(0)
    for data in polygons_data:
        data[:, ] -= min_coords
        data[:, ] /= (max_coords - min_coords)

    polygons = []
    for data in polygons_data:
        polygons.append(load_wkt(
            "POLYGON((%s))" % ",".join(" ".join(map(str, x)) for x in data)))

    return tuple(polygons)


# ----------------------

letter_polygons = {k: standardize_polygons_str(v) for k, v in all_letters.items()}

VOCABS = {"DNA": OrderedDict([("A", "green"),
                              ("C", "blue"),
                              ("G", "orange"),
                              ("T", "red")]),
          "RNA": OrderedDict([("A", "green"),
                              ("C", "blue"),
                              ("G", "orange"),
                              ("U", "red")]),
          "AA": OrderedDict([('A', '#CCFF00'),
                             ('B', "orange"),
                             ('C', '#FFFF00'),
                             ('D', '#FF0000'),
                             ('E', '#FF0066'),
                             ('F', '#00FF66'),
                             ('G', '#FF9900'),
                             ('H', '#0066FF'),
                             ('I', '#66FF00'),
                             ('K', '#6600FF'),
                             ('L', '#33FF00'),
                             ('M', '#00FF00'),
                             ('N', '#CC00FF'),
                             ('P', '#FFCC00'),
                             ('Q', '#FF00CC'),
                             ('R', '#0000FF'),
                             ('S', '#FF3300'),
                             ('T', '#FF6600'),
                             ('V', '#99FF00'),
                             ('W', '#00CCFF'),
                             ('Y', '#00FFCC'),
                             ('Z', 'blue')]),
          "RNAStruct": OrderedDict([("P", "red"),
                                    ("H", "green"),
                                    ("I", "blue"),
                                    ("M", "orange"),
                                    ("E", "violet")]),
          }

# vocabularies:
DNA = ["A", "C", "G", "T"]
RNA = ["A", "C", "G", "U"]
AMINO_ACIDS = ["A", "R", "N", "D", "B", "C", "E", "Q", "Z", "G", "H",
               "I", "L", "K", "M", "F", "P", "S", "T", "W", "Y", "V"]

# make sure things are in order
VOCABS["AA"] = OrderedDict((k, VOCABS["AA"][k]) for k in AMINO_ACIDS)
VOCABS["DNA"] = OrderedDict((k, VOCABS["DNA"][k]) for k in DNA)
VOCABS["RNA"] = OrderedDict((k, VOCABS["RNA"][k]) for k in RNA)


def add_letter_to_axis(ax, let, col, x, y, height):
    """Add 'let' with position x,y and height height to matplotlib axis 'ax'.
    """
    from shapely import affinity
    from descartes.patch import PolygonPatch
    if len(let) == 2:
        colors = [col, "white"]
    elif len(let) == 1:
        colors = [col]
    else:
        raise ValueError("3 or more Polygons are not supported")

    for polygon, color in zip(let, colors):
        new_polygon = affinity.scale(
            polygon, yfact=height, origin=(0, 0, 0))
        new_polygon = affinity.translate(
            new_polygon, xoff=x, yoff=y)
        patch = PolygonPatch(
            new_polygon, edgecolor=color, facecolor=color)
        ax.add_patch(patch)
    return



def encodeDNA(seq_vec, maxlen=None, seq_align="start"):
    """Convert the DNA sequence into 1-hot-encoding numpy array
    # Arguments
        seq_vec: list of chars
            List of sequences that can have different lengths
        maxlen: int or None,
            Should we trim (subset) the resulting sequence. If None don't trim.
            Note that trims wrt the align parameter.
            It should be smaller than the longest sequence.
        seq_align: character; 'end' or 'start'
            To which end should we align sequences?
    # Returns
        3D numpy array of shape (len(seq_vec), trim_seq_len(or maximal sequence length if None), 4)
    # Example
        ```python
            >>> sequence_vec = ['CTTACTCAGA', 'TCTTTA']
            >>> X_seq = encodeDNA(sequence_vec, seq_align="end", maxlen=8)
            >>> X_seq.shape
            (2, 8, 4)
            >>> print(X_seq)
            [[[0 0 0 1]
              [1 0 0 0]
              [0 1 0 0]
              [0 0 0 1]
              [0 1 0 0]
              [1 0 0 0]
              [0 0 1 0]
              [1 0 0 0]]
             [[0 0 0 0]
              [0 0 0 0]
              [0 0 0 1]
              [0 1 0 0]
              [0 0 0 1]
              [0 0 0 1]
              [0 0 0 1]
              [1 0 0 0]]]
        ```
    """
    return encodeSequence(seq_vec,
                          vocab=DNA,
                          neutral_vocab="N",
                          maxlen=maxlen,
                          seq_align=seq_align,
                          pad_value="N",
                          encode_type="one_hot")

def encodeSequence(seq_vec, vocab, neutral_vocab, maxlen=None,
                   seq_align="start", pad_value="N", encode_type="one_hot"):
    """Convert a list of genetic sequences into one-hot-encoded array.
    # Arguments
       seq_vec: list of strings (genetic sequences)
       vocab: list of chars: List of "words" to use as the vocabulary. Can be strings of length>0,
            but all need to have the same length. For DNA, this is: ["A", "C", "G", "T"].
       neutral_vocab: list of chars: Values used to pad the sequence or represent unknown-values. For DNA, this is: ["N"].
       maxlen: int or None,
            Should we trim (subset) the resulting sequence. If None don't trim.
            Note that trims wrt the align parameter.
            It should be smaller than the longest sequence.
       seq_align: character; 'end' or 'start'
            To which end should we align sequences?
       encode_type: "one_hot" or "token". "token" represents each vocab element as a positive integer from 1 to len(vocab) + 1.
                  neutral_vocab is represented with 0.
    # Returns
        Array with shape for encode_type:
            - "one_hot": `(len(seq_vec), maxlen, len(vocab))`
            - "token": `(len(seq_vec), maxlen)`
        If `maxlen=None`, it gets the value of the longest sequence length from `seq_vec`.
    """
    if isinstance(neutral_vocab, str):
        neutral_vocab = [neutral_vocab]
    if isinstance(seq_vec, str):
        raise ValueError("seq_vec should be an iterable returning " +
                         "strings not a string itself")
    assert len(vocab[0]) == len(pad_value)
    assert pad_value in neutral_vocab

    assert encode_type in ["one_hot", "token"]

    seq_vec = pad_sequences(seq_vec, maxlen=maxlen,
                            align=seq_align, value=pad_value)

    if encode_type == "one_hot":
        arr_list = [token2one_hot(tokenize(seq, vocab, neutral_vocab), len(vocab))
                    for i, seq in enumerate(seq_vec)]
    elif encode_type == "token":
        arr_list = [1 + np.array(tokenize(seq, vocab, neutral_vocab)) for seq in seq_vec]
        # we add 1 to be compatible with keras: https://keras.io/layers/embeddings/
        # indexes > 0, 0 = padding element

    return np.stack(arr_list)


def _get_vocab_dict(vocab):
    return {l: i for i, l in enumerate(vocab)}


def tokenize(seq, vocab, neutral_vocab=[]):
    """Convert sequence to integers
    # Arguments
       seq: Sequence to encode
       vocab: Vocabulary to use
       neutral_vocab: Neutral vocabulary -> assign those values to -1
    # Returns
       List of length `len(seq)` with integers from `-1` to `len(vocab) - 1`
    """
    # Req: all vocabs have the same length
    if isinstance(neutral_vocab, str):
        neutral_vocab = [neutral_vocab]

    nchar = len(vocab[0])
    for l in vocab + neutral_vocab:
        assert len(l) == nchar
    assert len(seq) % nchar == 0  # since we are using striding

    vocab_dict = _get_vocab_dict(vocab)
    for l in neutral_vocab:
        vocab_dict[l] = -1

    # current performance bottleneck
    return [vocab_dict[seq[(i * nchar):((i + 1) * nchar)]] for i in range(len(seq) // nchar)]



def token2one_hot(tvec, vocab_size):
    """
    Note: everything out of the vucabulary is transformed into `np.zeros(vocab_size)`
    """
    arr = np.zeros((len(tvec), vocab_size))

    tvec_range = np.arange(len(tvec))
    tvec = np.asarray(tvec)
    arr[tvec_range[tvec >= 0], tvec[tvec >= 0]] = 1
    return arr

def pad_sequences(sequence_vec, maxlen=None, align="end", value="N"):
    """Pad and/or trim a list of sequences to have common length. Procedure:
        1. Pad the sequence with N's or any other string or list element (`value`)
        2. Subset the sequence
    # Note
        See also: https://keras.io/preprocessing/sequence/
        Aplicable also for lists of characters
    # Arguments
        sequence_vec: list of chars or lists
            List of sequences that can have various lengths
        value: Neutral element to pad the sequence with. Can be `str` or `list`.
        maxlen: int or None; Final lenght of sequences.
             If None, maxlen is set to the longest sequence length.
        align: character; 'start', 'end' or 'center'
            To which end to align the sequences when triming/padding. See examples bellow.
    # Returns
        List of sequences of the same class as sequence_vec
    # Example
        ```python
            >>> sequence_vec = ['CTTACTCAGA', 'TCTTTA']
            >>> pad_sequences(sequence_vec, 10, align="start", value="N")
            ['CTTACTCAGA', 'TCTTTANNNN']
            >>> pad_sequences(sequence_vec, 10, align="end", value="N")
            ['CTTACTCAGA', 'NNNNTCTTTA']
            >>> pad_sequences(sequence_vec, 4, align="center", value="N")
            ['ACTC', 'CTTT']
        ```
    """

    # neutral element type checking
    assert isinstance(value, list) or isinstance(value, str)
    assert isinstance(value, type(sequence_vec[0]))
    assert not isinstance(sequence_vec, str)
    assert isinstance(sequence_vec[0], list) or isinstance(sequence_vec[0], str)

    max_seq_len = max([len(seq) for seq in sequence_vec])

    if maxlen is None:
        maxlen = max_seq_len
    else:
        maxlen = int(maxlen)

    if max_seq_len < maxlen:
        print("WARNING: Maximum sequence length (%s) is less than maxlen (%s)" % (max_seq_len, maxlen))
        max_seq_len = maxlen

    # check the case when len > 1
    for seq in sequence_vec:
        if not len(seq) % len(value) == 0:
            raise ValueError("All sequences need to be dividable by len(value)")
    if not maxlen % len(value) == 0:
        raise ValueError("maxlen needs to be dividable by len(value)")

    # pad and subset
    def pad(seq, max_seq_len, value="N", align="end"):
        seq_len = len(seq)
        assert max_seq_len >= seq_len
        if align is "end":
            n_left = max_seq_len - seq_len
            n_right = 0
        elif align is "start":
            n_right = max_seq_len - seq_len
            n_left = 0
        elif align is "center":
            n_left = (max_seq_len - seq_len) // 2 + (max_seq_len - seq_len) % 2
            n_right = (max_seq_len - seq_len) // 2
        else:
            raise ValueError("align can be of: end, start or center")

        # normalize for the length
        n_left = n_left // len(value)
        n_right = n_right // len(value)

        return value * n_left + seq + value * n_right

    def trim(seq, maxlen, align="end"):
        seq_len = len(seq)

        assert maxlen <= seq_len
        if align is "end":
            return seq[-maxlen:]
        elif align is "start":
            return seq[0:maxlen]
        elif align is "center":
            dl = seq_len - maxlen
            n_left = dl // 2 + dl % 2
            n_right = seq_len - dl // 2
            return seq[n_left:n_right]
        else:
            raise ValueError("align can be of: end, start or center")

    padded_sequence_vec = [pad(seq, max(max_seq_len, maxlen),
                               value=value, align=align) for seq in sequence_vec]
    padded_sequence_vec = [trim(seq, maxlen, align=align) for seq in padded_sequence_vec]

    return padded_sequence_vec

# TODO - add figsize???
def seqlogo(letter_heights, vocab="DNA", ax=None):
    """Make a logo plot
    # Arguments
        letter_heights: "motif length" x "vocabulary size" numpy array
    Can also contain negative values.
        vocab: str, Vocabulary name. Can be: DNA, RNA, AA, RNAStruct.
        ax: matplotlib axis
    """
    import matplotlib.pyplot as plt
    ax = ax or plt.gca()

    assert letter_heights.shape[1] == len(VOCABS[vocab])
    x_range = [1, letter_heights.shape[0]]
    pos_heights = np.copy(letter_heights)
    pos_heights[letter_heights < 0] = 0
    neg_heights = np.copy(letter_heights)
    neg_heights[letter_heights > 0] = 0

    for x_pos, heights in enumerate(letter_heights):
        letters_and_heights = sorted(zip(heights, list(VOCABS[vocab].keys())))
        y_pos_pos = 0.0
        y_neg_pos = 0.0
        for height, letter in letters_and_heights:
            color = VOCABS[vocab][letter]
            polygons = letter_polygons[letter]
            if height > 0:
                add_letter_to_axis(ax, polygons, color, 0.5 + x_pos, y_pos_pos, height)
                y_pos_pos += height
            else:
                add_letter_to_axis(ax, polygons, color, 0.5 + x_pos, y_neg_pos, height)
                y_neg_pos += height

    # if add_hline:
    #     ax.axhline(color="black", linewidth=1)
    ax.set_xlim(x_range[0] - 1, x_range[1] + 1)
    ax.grid(False)
    ax.set_xticks(list(range(*x_range)) + [x_range[-1]])
    ax.set_aspect(aspect='auto', adjustable='box')
    ax.autoscale_view()

def seqlogo_fig(letter_heights, vocab="DNA", figsize=(10, 2), ncol=1, plot_name=None):
    """
    # Arguments
        plot_name: Title of the plot. Can be a list of names
    """
    import matplotlib.pyplot as plt
    import math
    fig = plt.figure(figsize=figsize)

    if len(letter_heights.shape) == 3:
        #
        n_plots = letter_heights.shape[2]
        nrow = math.ceil(n_plots / ncol)
        if isinstance(plot_name, list):
            assert len(plot_name) == n_plots
    else:
        n_plots = 1
        nrow = 1
        ncol = 1

    for i in range(n_plots):
        if len(letter_heights.shape) == 3:
            w_cur = letter_heights[:, :, i]
        else:
            w_cur = letter_heights
        ax = plt.subplot(nrow, ncol, i + 1)
        plt.tight_layout()

        # plot the motif
        seqlogo(w_cur, vocab, ax)

        # add the title
        if plot_name is not None:
            if n_plots > 0:
                if isinstance(plot_name, list):
                    pln = plot_name[i]
                else:
                    pln = plot_name + " {0}".format(i)
            else:
                pln = plot_name
            ax.set_title(pln)
    return fig


####### copy end


"""Postprocessing CLI
"""
from __future__ import absolute_import
from __future__ import print_function

import argparse
import logging
import os
import sys

import kipoi
from kipoi.cli.parser_utils import add_model, add_dataloader, file_exists, dir_exists
from kipoi.postprocessing.variant_effects.components import VarEffectFuncType
from kipoi.postprocessing.variant_effects.scores import get_scoring_fns
from kipoi import writers
from kipoi.utils import cd
from kipoi.utils import parse_json_file_str

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _prepare_multi_model_args(args):
    assert isinstance(args.model, list)
    assert isinstance(args.source, list)
    assert isinstance(args.seq_length, list)
    assert isinstance(args.dataloader, list)
    assert isinstance(args.dataloader_source, list)
    assert isinstance(args.dataloader_args, list)

    def ensure_matching_args(ref_arg, query_arg, ref_label, query_label, allow_zero=True):
        assert isinstance(ref_arg, list)
        assert isinstance(query_arg, list)
        n = len(ref_arg)
        if allow_zero and (len(query_arg) == 0):
            ret = [None] * n
        elif len(query_arg) == 1:
            ret = [query_arg[0]] * n
        elif not len(query_arg) == n:
            raise Exception("Either give one {q} for all {r} or one {q} for every {r} in the same order.".format(
                q=query_label, r=ref_label))
        else:
            ret = query_arg
        return ret

    args.source = ensure_matching_args(args.model, args.source, "--model", "--source", allow_zero=False)
    args.seq_length = ensure_matching_args(args.model, args.seq_length, "--model", "--seq_length")
    args.dataloader = ensure_matching_args(args.model, args.dataloader, "--model", "--dataloader")
    args.dataloader_source = ensure_matching_args(args.dataloader, args.dataloader_source, "--dataloader",
                                                  "--dataloader_source")
    args.dataloader_args = ensure_matching_args(args.model, args.dataloader_args, "--model",
                                                "--dataloader_args", allow_zero=False)


def cli_score_variants(command, raw_args):
    """CLI interface to score variants
    """
    # Updated argument names:
    # - scoring -> scores
    # - --vcf_path -> --input_vcf, -i
    # - --out_vcf_fpath -> --output_vcf, -o
    # - --output -> -e, --extra_output
    # - remove - -install_req
    # - scoring_kwargs -> score_kwargs
    AVAILABLE_FORMATS = ["tsv", "hdf5", "h5"]
    assert command == "score_variants"
    parser = argparse.ArgumentParser('kipoi postproc {}'.format(command),
                                     description='Predict effect of SNVs using ISM.')
    parser.add_argument('model', help='Model name.', nargs="+")
    parser.add_argument('--source', default=["kipoi"], nargs="+",
                        choices=list(kipoi.config.model_sources().keys()),
                        help='Model source to use. Specified in ~/.kipoi/config.yaml' +
                             " under model_sources. " +
                             "'dir' is an additional source referring to the local folder.")
    parser.add_argument('--dataloader', nargs="+", default=[],
                        help="Dataloader name. If not specified, the model's default" +
                             "DataLoader will be used")
    parser.add_argument('--dataloader_source', nargs="+", default=["kipoi"],
                        help="Dataloader source")
    parser.add_argument('--dataloader_args', nargs="+", default=[],
                        help="Dataloader arguments either as a json string:" +
                             "'{\"arg1\": 1} or as a file path to a json file")
    parser.add_argument('-i', '--input_vcf',
                        help='Input VCF.')
    parser.add_argument('-o', '--output_vcf',
                        help='Output annotated VCF file path.', default=None)
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument('-r', '--restriction_bed', default=None,
                        help="Regions for prediction can only be subsets of this bed file")
    parser.add_argument('-e', '--extra_output', required=False,
                        help="Additional output file. File format is inferred from the file path ending" +
                             ". Available file formats are: {0}".format(",".join(AVAILABLE_FORMATS)))
    parser.add_argument('-s', "--scores", default="diff", nargs="+",
                        help="Scoring method to be used. Only scoring methods selected in the model yaml file are"
                             "available except for `diff` which is always available. Select scoring function by the"
                             "`name` tag defined in the model yaml file.")
    parser.add_argument('-k', "--score_kwargs", default="", nargs="+",
                        help="JSON definition of the kwargs for the scoring functions selected in --scoring. The "
                             "definiton can either be in JSON in the command line or the path of a .json file. The "
                             "individual JSONs are expected to be supplied in the same order as the labels defined in "
                             "--scoring. If the defaults or no arguments should be used define '{}' for that respective "
                             "scoring method.")
    parser.add_argument('-l', "--seq_length", type=int, nargs="+", default=[],
                        help="Optional parameter: Model input sequence length - necessary if the model does not have a "
                             "pre-defined input sequence length.")
    parser.add_argument('--std_var_id', action="store_true", help="If set then variant IDs in the annotated"
                                                                  " VCF will be replaced with a standardised, unique ID.")

    args = parser.parse_args(raw_args)
    # Make sure all the multi-model arguments like source, dataloader etc. fit together
    _prepare_multi_model_args(args)

    # Check that all the folders exist
    file_exists(args.input_vcf, logger)
    dir_exists(os.path.dirname(args.output_vcf), logger)
    if args.extra_output is not None:
        dir_exists(os.path.dirname(args.extra_output), logger)

        # infer the file format
        args.file_format = args.extra_output.split(".")[-1]
        if args.file_format not in AVAILABLE_FORMATS:
            logger.error("File ending: {0} for file {1} not from {2}".
                         format(args.file_format, args.extra_output, AVAILABLE_FORMATS))
            sys.exit(1)

        if args.file_format in ["hdf5", "h5"]:
            # only if hdf5 output is used
            import deepdish

    if not isinstance(args.scores, list):
        args.scores = [args.scores]

    score_kwargs = []
    if len(args.score_kwargs) > 0:
        score_kwargs = args.score_kwargs
        if len(args.scores) >= 1:
            # Check if all scoring functions should be used:
            if args.scores == ["all"]:
                if len(score_kwargs) >= 1:
                    raise ValueError("`--score_kwargs` cannot be defined in combination will `--scoring all`!")
            else:
                score_kwargs = [parse_json_file_str(el) for el in score_kwargs]
                if not len(args.score_kwargs) == len(score_kwargs):
                    raise ValueError("When defining `--score_kwargs` a JSON representation of arguments (or the "
                                     "path of a file containing them) must be given for every "
                                     "`--scores` function.")

    keep_predictions = args.extra_output is not None

    n_models = len(args.model)

    res = {}
    for model_name, model_source, dataloader, dataloader_source, dataloader_args, seq_length in zip(args.model,
                                                                                                    args.source,
                                                                                                    args.dataloader,
                                                                                                    args.dataloader_source,
                                                                                                    args.dataloader_args,
                                                                                                    args.seq_length):
        model_name_safe = model_name.replace("/", "_")
        output_vcf_model = None
        if args.output_vcf is not None:
            output_vcf_model = args.output_vcf
            # If multiple models are to be analysed then vcfs need renaming.
            if n_models > 1:
                if output_vcf_model.endswith(".vcf"):
                    output_vcf_model = output_vcf_model[:-4]
                output_vcf_model += model_name_safe + ".vcf"

        dataloader_arguments = parse_json_file_str(dataloader_args)

        # --------------------------------------------
        # load model & dataloader
        model = kipoi.get_model(model_name, model_source)

        if dataloader is not None:
            Dl = kipoi.get_dataloader_factory(dataloader, dataloader_source)
        else:
            Dl = model.default_dataloader

        # Load effect prediction related model info
        model_info = kipoi.postprocessing.variant_effects.ModelInfoExtractor(model, Dl)

        if model_info.use_seq_only_rc:
            logger.info('Model SUPPORTS simple reverse complementation of input DNA sequences.')
        else:
            logger.info('Model DOES NOT support simple reverse complementation of input DNA sequences.')

        if output_vcf_model is not None:
            logger.info('Annotated VCF will be written to %s.' % str(output_vcf_model))

        res[model_name_safe] = kipoi.postprocessing.variant_effects.score_variants(model,
                                                                                   dataloader_arguments,
                                                                                   args.input_vcf,
                                                                                   output_vcf_model,
                                                                                   scores=args.scores,
                                                                                   score_kwargs=score_kwargs,
                                                                                   num_workers=args.num_workers,
                                                                                   batch_size=args.batch_size,
                                                                                   seq_length=seq_length,
                                                                                   std_var_id=args.std_var_id,
                                                                                   restriction_bed=args.restriction_bed,
                                                                                   return_predictions=keep_predictions)

    # tabular files
    if keep_predictions:
        if args.file_format in ["tsv"]:
            for model_name in res:
                for i, k in enumerate(res[model_name]):
                    # Remove an old file if it is still there...
                    if i == 0:
                        try:
                            os.unlink(args.extra_output)
                        except Exception:
                            pass
                    with open(args.extra_output, "w") as ofh:
                        ofh.write("KPVEP_%s:%s\n" % (k.upper(), model_name))
                        res[model_name][k].to_csv(args.extra_output, sep="\t", mode="a")

        if args.file_format in ["hdf5", "h5"]:
            deepdish.io.save(args.extra_output, res)

    logger.info('Successfully predicted samples')


def isint(qstr):
    import re
    return bool(re.match("^[0-9]+$", qstr))


# Parse the slice
def parse_filter_slice(in_str):
    if in_str.startswith("(") or in_str.startswith("["):
        in_str_els = in_str.lstrip("([").rstrip(")]").split(",")
        slices = []
        for slice_el in in_str_els:
            slice_el = slice_el.strip(" ")
            if slice_el == "...":
                slices.append(Ellipsis)
            elif isint(slice_el):
                slices.append(int(slice_el))
                if len(in_str_els) == 1:
                    return int(slice_el)
            else:
                # taken from https://stackoverflow.com/questions/680826/python-create-slice-object-from-string
                slices.append(slice(*map(lambda x: int(x.strip()) if x.strip() else None, slice_el.split(':'))))
        return tuple(slices)
    elif isint(in_str):
        return int(in_str)
    else:
        raise Exception("Filter index slice not valid. Allowed values are e.g.: '1', [1:3,...], [:, 0:4, ...]")


def cli_grad(command, raw_args):
    """CLI interface to predict
    """
    from .main import prepare_batch
    from kipoi.model import GradientMixin
    assert command == "grad"
    from tqdm import tqdm
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Save gradients and inputs to a hdf5 file.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument("-i", "--install_req", action='store_true',
                        help="Install required packages from requirements.txt")
    parser.add_argument("-l", "--layer", default=None,
                        help="Which output layer to use to make the predictions. If specified," +
                             "`model.predict_activation_on_batch` will be invoked instead of `model.predict_on_batch`",
                        required=False)
    parser.add_argument("--final_layer",
                        help="Alternatively to `--layer` this flag can be used to indicate that the last layer should "
                             "be used.", action='store_true')
    parser.add_argument("--pre_nonlinearity",
                        help="Flag indicating that it should checked whether the selected output is post activation "
                             "function. If a non-linear activation function is used attempt to use its input. This "
                             "feature is not available for all models.", action='store_true')
    parser.add_argument("-f", "--filter_ind",
                        help="Filter index that should be inspected with gradients. If not set all filters will " +
                             "be used.", default=None)
    parser.add_argument("-a", "--avg_func",
                        help="Averaging function to be applied across selected filters (`--filter_ind`) in " +
                             "layer `--layer`.", choices=GradientMixin.allowed_functions, default="sum")
    parser.add_argument('--selected_fwd_node', help="If the selected layer has multiple inbound connections in "
                                                    "the graph then those can be selected here with an integer "
                                                    "index. Not necessarily supported by all models.",
                        default=None, type=int)
    parser.add_argument('-o', '--output', required=True, nargs="+",
                        help="Output files. File format is inferred from the file path ending. Available file formats are: " +
                             ", ".join(["." + k for k in writers.FILE_SUFFIX_MAP]))
    args = parser.parse_args(raw_args)

    dataloader_kwargs = parse_json_file_str(args.dataloader_args)

    # setup the files
    if not isinstance(args.output, list):
        args.output = [args.output]
    for o in args.output:
        ending = o.split('.')[-1]
        if ending not in writers.FILE_SUFFIX_MAP:
            logger.error("File ending: {0} for file {1} not from {2}".
                         format(ending, o, writers.FILE_SUFFIX_MAP))
            sys.exit(1)
        dir_exists(os.path.dirname(o), logger)
    # --------------------------------------------
    # install args
    if args.install_req:
        kipoi.pipeline.install_model_requirements(args.model,
                                                  args.source,
                                                  and_dataloaders=True)

    layer = args.layer
    if layer is None and not args.final_layer:
        raise Exception("A layer has to be selected explicitely using `--layer` or implicitely by using the"
                        "`--final_layer` flag.")

    # Not a good idea
    # if layer is not None and isint(layer):
    #    logger.warn("Interpreting `--layer` value as integer layer index!")
    #    layer = int(args.layer)


    # load model & dataloader
    model = kipoi.get_model(args.model, args.source)

    if not isinstance(model, GradientMixin):
        raise Exception("Model does not support gradient calculation.")

    if args.dataloader is not None:
        Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
    else:
        Dl = model.default_dataloader

    dataloader_kwargs = kipoi.pipeline.validate_kwargs(Dl, dataloader_kwargs)
    dl = Dl(**dataloader_kwargs)

    filter_ind_parsed = None
    if args.filter_ind is not None:
        filter_ind_parsed = parse_filter_slice(args.filter_ind)

    # setup batching
    it = dl.batch_iter(batch_size=args.batch_size,
                       num_workers=args.num_workers)

    # Setup the writers
    use_writers = []
    for output in args.output:
        ending = output.split('.')[-1]
        W = writers.FILE_SUFFIX_MAP[ending]
        logger.info("Using {0} for file {1}".format(W.__name__, output))
        if ending == "tsv":
            assert W == writers.TsvBatchWriter
            use_writers.append(writers.TsvBatchWriter(file_path=output, nested_sep="/"))
        elif ending == "bed":
            raise Exception("Please use tsv or hdf5 output format.")
        elif ending in ["hdf5", "h5"]:
            assert W == writers.HDF5BatchWriter
            use_writers.append(writers.HDF5BatchWriter(file_path=output))
        else:
            logger.error("Unknown file format: {0}".format(ending))
            sys.exit(1)

    # Loop through the data, make predictions, save the output
    for i, batch in enumerate(tqdm(it)):
        # validate the data schema in the first iteration
        if i == 0 and not Dl.output_schema.compatible_with_batch(batch):
            logger.warn("First batch of data is not compatible with the dataloader schema.")

        # make the prediction
        pred_batch = model.input_grad(batch['inputs'], filter_idx=filter_ind_parsed,
                                      avg_func=args.avg_func, layer=layer, final_layer=args.final_layer,
                                      selected_fwd_node=args.selected_fwd_node,
                                      pre_nonlinearity=args.pre_nonlinearity)

        # write out the predictions, metadata (, inputs, targets)
        # always keep the inputs so that input*grad can be generated!
        output_batch = prepare_batch(batch, pred_batch, keep_inputs=True)
        for writer in use_writers:
            writer.batch_write(output_batch)

    for writer in use_writers:
        writer.close()
    logger.info('Done! Gradients stored in {0}'.format(",".join(args.output)))


def cli_grad_to_file(command, raw_args):
    """ CLI to save seq inputs of grad*input to a bigwig file
    """
    assert command == "gr_inp_to_file"
    parser = argparse.ArgumentParser('kipoi postproc {}'.format(command),
                                     description='Save grad*input in a file.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    # TODO - rename path to fpath
    parser.add_argument('-f', '--input_file', required=False,
                        help="Input HDF5 file produced from `grad`")
    parser.add_argument('-o', '--output', required=False,
                        help="Output bigwig for bedgraph file")
    parser.add_argument('--sample', required=False, type=int, default=None,
                        help="Input line for which the BigWig file should be generated. If not defined all"
                             "samples will be written.")
    parser.add_argument('--model_input', required=False, default=None,
                        help="Model input name to be used for plotting. As defined in model.yaml. Can be omitted if"
                             "model only has one input.")
    args = parser.parse_args(raw_args)

    # Check that all the folders exist
    dir_exists(os.path.dirname(args.output), logger)
    # --------------------------------------------
    # install args
    import matplotlib.pyplot
    matplotlib.pyplot.switch_backend('agg')
    import matplotlib.pylab as plt
    from kipoi.postprocessing.variant_effects.mutation_map import MutationMapPlotter
    from kipoi.postprocessing.gradient_vis.vis import GradPlotter
    from kipoi.writers import BedGraphWriter

    logger.info('Loading gradient results file and model info...')

    gp = GradPlotter.from_hdf5(args.input_file, model=args.model, source=args.source)

    if args.sample is not None:
        samples = [args.sample]
    else:
        samples = list(range(gp.get_num_samples(args.model_input)))

    if args.output.endswith(".bed") or args.output.endswith(".bedgraph"):
        of_obj = BedGraphWriter(args.output)
    else:
        raise Exception("Output file format not supported!")

    logger.info('Writing...')

    for sample in samples:
        gp.write(sample, model_input=args.model_input, writer_obj=of_obj)

    logger.info('Saving...')

    of_obj.close()

    logger.info('Successfully wrote grad*input to file.')


def cli_create_mutation_map(command, raw_args):
    """CLI interface to calculate mutation map data 
    """
    assert command == "create_mutation_map"
    parser = argparse.ArgumentParser('kipoi postproc {}'.format(command),
                                     description='Predict effect of SNVs using ISM.')
    add_model(parser)
    add_dataloader(parser, with_args=True)
    parser.add_argument('-r', '--regions_file',
                        help='Region definition as VCF or bed file. Not a required input.')
    # TODO - rename path to fpath
    parser.add_argument('--batch_size', type=int, default=32,
                        help='Batch size to use in prediction')
    parser.add_argument("-n", "--num_workers", type=int, default=0,
                        help="Number of parallel workers for loading the dataset")
    parser.add_argument("-i", "--install_req", action='store_true',
                        help="Install required packages from requirements.txt")
    parser.add_argument('-o', '--output', required=True,
                        help="Output HDF5 file. To be used as input for plotting.")
    parser.add_argument('-s', "--scores", default="diff", nargs="+",
                        help="Scoring method to be used. Only scoring methods selected in the model yaml file are"
                             "available except for `diff` which is always available. Select scoring function by the"
                             "`name` tag defined in the model yaml file.")
    parser.add_argument('-k', "--score_kwargs", default="", nargs="+",
                        help="JSON definition of the kwargs for the scoring functions selected in --scores. The "
                             "definiton can either be in JSON in the command line or the path of a .json file. The "
                             "individual JSONs are expected to be supplied in the same order as the labels defined in "
                             "--scores. If the defaults or no arguments should be used define '{}' for that respective "
                             "scoring method.")
    parser.add_argument('-l', "--seq_length", type=int, default=None,
                        help="Optional parameter: Model input sequence length - necessary if the model does not have a "
                             "pre-defined input sequence length.")

    args = parser.parse_args(raw_args)

    # extract args for kipoi.variant_effects.predict_snvs

    dataloader_arguments = parse_json_file_str(args.dataloader_args)

    if args.output is None:
        raise Exception("Output file `--output` has to be set!")

    # --------------------------------------------
    # install args
    if args.install_req:
        kipoi.pipeline.install_model_requirements(args.model, args.source, and_dataloaders=True)
    # load model & dataloader
    model = kipoi.get_model(args.model, args.source)

    regions_file = os.path.realpath(args.regions_file)
    output = os.path.realpath(args.output)
    with cd(model.source_dir):
        if not os.path.exists(regions_file):
            raise Exception("Regions inputs file does not exist: %s" % args.regions_file)

        # Check that all the folders exist
        file_exists(regions_file, logger)
        dir_exists(os.path.dirname(output), logger)

        if args.dataloader is not None:
            Dl = kipoi.get_dataloader_factory(args.dataloader, args.dataloader_source)
        else:
            Dl = model.default_dataloader

    if not isinstance(args.scores, list):
        args.scores = [args.scores]

    dts = get_scoring_fns(model, args.scores, args.score_kwargs)

    # Load effect prediction related model info
    model_info = kipoi.postprocessing.variant_effects.ModelInfoExtractor(model, Dl)
    manual_seq_len = args.seq_length

    # Select the appropriate region generator and vcf or bed file input
    args.file_format = regions_file.split(".")[-1]
    bed_region_file = None
    vcf_region_file = None
    bed_to_region = None
    vcf_to_region = None
    if args.file_format == "vcf" or regions_file.endswith("vcf.gz"):
        vcf_region_file = regions_file
        if model_info.requires_region_definition:
            # Select the SNV-centered region generator
            vcf_to_region = kipoi.postprocessing.variant_effects.SnvCenteredRg(model_info, seq_length=manual_seq_len)
            logger.info('Using variant-centered sequence generation.')
    elif args.file_format == "bed":
        if model_info.requires_region_definition:
            # Select the SNV-centered region generator
            bed_to_region = kipoi.postprocessing.variant_effects.BedOverlappingRg(model_info, seq_length=manual_seq_len)
            logger.info('Using bed-file based sequence generation.')
        bed_region_file = regions_file
    else:
        raise Exception("")

    if model_info.use_seq_only_rc:
        logger.info('Model SUPPORTS simple reverse complementation of input DNA sequences.')
    else:
        logger.info('Model DOES NOT support simple reverse complementation of input DNA sequences.')

    from kipoi.postprocessing.variant_effects.mutation_map import _generate_mutation_map
    mdmm = _generate_mutation_map(model,
                                  Dl,
                                  vcf_fpath=vcf_region_file,
                                  bed_fpath=bed_region_file,
                                  batch_size=args.batch_size,
                                  num_workers=args.num_workers,
                                  dataloader_args=dataloader_arguments,
                                  vcf_to_region=vcf_to_region,
                                  bed_to_region=bed_to_region,
                                  evaluation_function_kwargs={'diff_types': dts},
                                  )
    mdmm.save_to_file(output)

    logger.info('Successfully generated mutation map data')


def cli_plot_mutation_map(command, raw_args):
    """CLI interface to plot mutation map
    """
    assert command == "plot_mutation_map"
    parser = argparse.ArgumentParser('kipoi postproc {}'.format(command),
                                     description='Plot mutation map in a file.')
    # TODO - rename path to fpath
    parser.add_argument('-f', '--input_file', required=False,
                        help="Input HDF5 file produced from `create_mutation_map`")
    parser.add_argument('-o', '--output', required=False,
                        help="Output image file")
    parser.add_argument('--input_entry', required=True, type=int,
                        help="Input line for which the plot should be generated")
    parser.add_argument('--model_seq_input', required=True,
                        help="Model input name to be used for plotting. As defined in model.yaml.")
    parser.add_argument('--scoring_key', required=True,
                        help="Variant score label to be used for plotting. As defined when running "
                             "`create_mutation_map`.")
    parser.add_argument('--model_output', required=True,
                        help="Model output to be used for plotting. As defined in model.yaml.")
    parser.add_argument('--limit_region_genomic', required=False, nargs =2, type = int, default=None,
                        help="Limit to genomic region. Given as tuple without chromosome, "
                             "eg: `--limit_region_genomic 13245 12347`")
    parser.add_argument('--rc_plot', required=False, action="store_true",
                        help="Make reverse-complement plot.")
    args = parser.parse_args(raw_args)

    # Check that all the folders exist
    dir_exists(os.path.dirname(args.output), logger)
    # --------------------------------------------
    # install args
    import matplotlib.pyplot
    matplotlib.pyplot.switch_backend('agg')
    import matplotlib.pylab as plt
    from kipoi.postprocessing.variant_effects.mutation_map import MutationMapPlotter

    logger.info('Loading mutation map file...')

    mutmap = MutationMapPlotter(fname=args.input_file)

    fig = plt.figure(figsize=(20, 2))
    ax = plt.subplot(1, 1, 1)

    logger.info('Plotting...')

    if args.limit_region_genomic is not None:
        args.limit_region_genomic = tuple(args.limit_region_genomic)


    mutmap.plot_mutmap(args.input_entry, args.model_seq_input, args.scoring_key, args.model_output, ax=ax,
                       limit_region_genomic=args.limit_region_genomic, rc_plot=args.rc_plot)
    fig.savefig(args.output)

    logger.info('Successfully plotted mutation map')


# --------------------------------------------
# CLI commands


command_functions = {
    'score_variants': cli_score_variants,
    'grad': cli_grad,
    'create_mutation_map': cli_create_mutation_map,
    'plot_mutation_map': cli_plot_mutation_map,
    'gr_inp_to_file': cli_grad_to_file
}
commands_str = ', '.join(command_functions.keys())

parser = argparse.ArgumentParser(
    description='Kipoi model-zoo command line tool',
    usage='''kipoi postproc <command> [-h] ...

    # Available sub-commands:
    score_variants        Score variants with a kipoi model
    grad                  Save gradients and inputs to a hdf5 file
    gr_inp_to_file        Save grad*input in a file.
    create_mutation_map   Calculate variant effect scores for mutation map plotting
    plot_mutation_map     Plot mutation map from data generated in `create_mutation_map`
    ''')
parser.add_argument('command', help='Subcommand to run; possible commands: {}'.format(commands_str))


def cli_main(command, raw_args):
    args = parser.parse_args(raw_args[0:1])
    if args.command not in command_functions:
        parser.print_help()
        parser.exit(
            status=1,
            message='\nCommand `{}` not found. Possible commands: {}\n'.format(
                args.command, commands_str))
    command_fn = command_functions[args.command]
    command_fn(args.command, raw_args[1:])



from __future__ import absolute_import
from __future__ import print_function

import copy
import warnings
import numpy as np
import abc

import kipoi
import logging
import six
from kipoi.postprocessing.variant_effects.components import VarEffectFuncType
from kipoi.utils import load_module, getargs, parse_json_file_str

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def default_kwargs(args):
    """Return the example kwargs
    """
    return {k: v.default for k, v in six.iteritems(args) if v.default is not None}


class Score(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        raise NotImplementedError("Analysis routine has to be implemented")


class RCScore(Score):
    allowed_str_opts = ["min", "max", "mean", "median", "absmax"]

    #

    def __init__(self, rc_merging="mean"):
        if isinstance(rc_merging, str):
            if rc_merging == "absmax":
                self.rc_merging = self.absmax
            elif rc_merging in self.allowed_str_opts:
                self.rc_merging = lambda x, y: getattr(np, rc_merging)([x, y], axis=0)
        elif callable(rc_merging):
            self.rc_merging = rc_merging
        else:
            raise Exception("rc_merging has to be a callable function of a string: %s" % str(self.allowed_str_opts))

    #

    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        raise NotImplementedError("Analysis routine has to be implemented")

    @staticmethod
    def absmax(x, y, inplace=True):
        if not inplace:
            x = copy.deepcopy(x)
        replace_filt = np.abs(x) < np.abs(y)
        x[replace_filt] = y[replace_filt]
        return x


class Logit(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        preds = {"ref": ref, "ref_rc": ref_rc, "alt": alt, "alt_rc": alt_rc}
        if np.any([(preds[k].min() < 0 or preds[k].max() > 1) for k in preds if preds[k] is not None]):
            warnings.warn("Using log_odds on model outputs that are not bound [0,1]")
        diffs = np.log(preds["alt"] / (1 - preds["alt"])) - np.log(preds["ref"] / (1 - preds["ref"]))

        if (preds["ref_rc"] is not None) and ((preds["alt_rc"] is not None)):
            diffs_rc = np.log(preds["alt_rc"] / (1 - preds["alt_rc"])) - np.log(preds["ref_rc"] / (1 - preds["ref_rc"]))
            return self.rc_merging(diffs, diffs_rc)
        else:
            return diffs


class LogitAlt(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        preds = {"ref": ref, "ref_rc": ref_rc, "alt": alt, "alt_rc": alt_rc}
        if np.any([(preds[k].min() < 0 or preds[k].max() > 1) for k in preds if preds[k] is not None]):
            warnings.warn("Using log_odds on model outputs that are not bound [0,1]")
        logits = np.log(preds["alt"] / (1 - preds["alt"]))

        if preds["alt_rc"] is not None:
            logits_rc = np.log(preds["alt_rc"] / (1 - preds["alt_rc"]))
            return self.rc_merging(logits, logits_rc)
        else:
            return logits


class LogitRef(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        preds = {"ref": ref, "ref_rc": ref_rc, "alt": alt, "alt_rc": alt_rc}
        if np.any([(preds[k].min() < 0 or preds[k].max() > 1) for k in preds if preds[k] is not None]):
            warnings.warn("Using log_odds on model outputs that are not bound [0,1]")
        logits = np.log(preds["ref"] / (1 - preds["ref"]))

        if preds["ref_rc"] is not None:
            logits_rc = np.log(preds["ref_rc"] / (1 - preds["ref_rc"]))
            return self.rc_merging(logits, logits_rc)
        else:
            return logits


class Alt(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        alt_out = alt
        if alt_rc is not None:
            alt_out = self.rc_merging(alt, alt_rc)
        return alt_out


class Ref(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        ref_out = ref
        if ref_rc is not None:
            ref_out = self.rc_merging(ref, ref_rc)
        return ref_out


class Diff(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        preds = {"ref": ref, "ref_rc": ref_rc, "alt": alt, "alt_rc": alt_rc}
        diffs = preds["alt"] - preds["ref"]

        if (preds["ref_rc"] is not None) and ((preds["alt_rc"] is not None)):
            diffs_rc = preds["alt_rc"] - preds["ref_rc"]
            return self.rc_merging(diffs, diffs_rc)
        else:
            return diffs


class DeepSEA_effect(RCScore):
    def __call__(self, ref, alt, ref_rc=None, alt_rc=None):
        preds = {"ref": ref, "ref_rc": ref_rc, "alt": alt, "alt_rc": alt_rc}
        if np.any([(preds[k].min() < 0 or preds[k].max() > 1) for k in preds if preds[k] is not None]):
            warnings.warn("Using log_odds on model outputs that are not bound [0,1]")
        logit_diffs = np.log(preds["alt"] / (1 - preds["alt"])) - np.log(preds["ref"] / (1 - preds["ref"]))
        diffs = preds["alt"] - preds["ref"]

        if (preds["ref_rc"] is not None) and ((preds["alt_rc"] is not None)):
            logit_diffs_rc = np.log(preds["alt_rc"] / (1 - preds["alt_rc"])) - np.log(
                preds["ref_rc"] / (1 - preds["ref_rc"]))
            diffs_rc = preds["alt_rc"] - preds["ref_rc"]
            logit_diffs = self.rc_merging(logit_diffs, logit_diffs_rc)
            diffs = self.rc_merging(diffs, diffs_rc)
            # self.rc_merging(np.abs(logit_diffs) * np.abs(diffs), np.abs(logit_diffs_rc) * np.abs(diffs_rc))

        return np.abs(logit_diffs) * np.abs(diffs)


# if e.g. 'logit' is allowed then also 'logit_alt' is allowed, etc. values here refer to 'scoring_options' keys.
categorical_enable = {
    "__any__": ["diff", "ref", "alt"],
    "logit": ["logit", "logit_ref", "logit_alt", "deepsea_effect"]
}
builtin_default_kwargs = {"rc_merging": "mean"}

scoring_options = {
    # deepsea_effect diff logit_diff logit_ref logit_alt
    # TODO - we should add more options to it: ref, alt, ref_rc, alt_rc
    "ref": Ref,
    "alt": Alt,
    "diff": Diff,
    "logit_ref": LogitRef,
    "logit_alt": LogitAlt,
    "logit": Logit,
    "deepsea_effect": DeepSEA_effect
}

scoring_names = {
    VarEffectFuncType.diff: "diff",
    VarEffectFuncType.ref: "ref",
    VarEffectFuncType.alt: "alt",
    VarEffectFuncType.logit: "logit",
    VarEffectFuncType.logit_ref: "logit_ref",
    VarEffectFuncType.logit_alt: "logit_alt",
    VarEffectFuncType.deepsea_effect: "deepsea_effect",
}


def get_avail_scoring_fns(model):
    if model.postprocessing.variant_effects is None:
        raise Exception("Model deosn't support variant effect prediction according model yaml file.")
    avail_scoring_fns = []  # contains callables
    avail_scoring_fn_def_args = []  # default kwargs for every callable
    avail_scoring_fn_names = []  # contains the labels
    default_scoring_fns = []  # contains the labels of the defaults
    for sf in model.postprocessing.variant_effects.scoring_functions:
        if sf.type is not VarEffectFuncType.custom:
            sn = scoring_names[sf.type]
            sf_obj = scoring_options[sn]
            s_label = sn
            if (sf.name != "") and (sf.name not in scoring_options):
                if sf.name in avail_scoring_fn_names:
                    raise Exception("Scoring function names have to unique in the model yaml file.")
                s_label = sf.name
            def_kwargs = builtin_default_kwargs
        else:
            prefix = ""
            if (sf.name == "") or (sf.name in scoring_options):
                prefix = "custom_"
            s_label = prefix + sf.name
            if s_label in avail_scoring_fn_names:
                raise Exception("Scoring function names have to unique in the model yaml file.")
            if sf.defined_as == "":
                raise Exception("`defined_as` has to be defined for a custom function.")
            file_path, obj_name = tuple(sf.defined_as.split("::"))
            sf_obj = getattr(load_module(file_path), obj_name)
            # check that the scoring function arguments match yaml arguments
            if not getargs(sf_obj) == set(sf.args.keys()):
                raise ValueError("Scoring function arguments: \n{0}\n don't match ".format(set(getargs(sf_obj))) +
                                 "the specification in the dataloader.yaml file:\n{0}".
                                 format(set(sf.args.keys())))

            def_kwargs = None
            if all([(sf.args[k].default is not None) or (sf.args[k].optional) for k in sf.args]):
                def_kwargs = default_kwargs(sf.args)
            if len(sf.args) == 0:
                def_kwargs = {}  # indicates that this callable doesn't accept any arguments for initialisation.

        if s_label in avail_scoring_fn_names:
            raise Exception("Mulitple scoring functions defined with name '%s' in the model yaml file!" % s_label)

        avail_scoring_fn_def_args.append(def_kwargs)
        avail_scoring_fns.append(sf_obj)
        avail_scoring_fn_names.append(s_label)
        if sf.default:
            default_scoring_fns.append(s_label)

    # if no default scoring functions have been set then take all of the above.
    if len(default_scoring_fns) == 0:
        default_scoring_fns = copy.copy(avail_scoring_fn_names)

    # try to complete the set of functions if needed. None of the additional ones will become part of the defaults
    additional_scoring_fn_def_args = []
    additional_scoring_fns = []
    additional_scoring_fn_names = []
    for scoring_fn, def_args in zip(avail_scoring_fns, avail_scoring_fn_def_args):
        scoring_name = None
        for sn in scoring_options:
            if scoring_fn is scoring_options[sn]:
                scoring_name = sn
                break
        if scoring_name is None:
            continue
        categories = [cat for cat in categorical_enable if scoring_name in categorical_enable[cat]]
        for cat in categories:
            for scoring_name in categorical_enable[cat]:
                if (scoring_options[scoring_name] not in avail_scoring_fns) and \
                        (scoring_options[scoring_name] not in additional_scoring_fns):
                    additional_scoring_fn_def_args.append(def_args)
                    additional_scoring_fns.append(scoring_options[scoring_name])
                    s_label = scoring_name
                    if s_label in avail_scoring_fn_names:
                        s_label = "default_" + s_label
                    additional_scoring_fn_names.append(s_label)

    avail_scoring_fns += additional_scoring_fns
    avail_scoring_fn_def_args += additional_scoring_fn_def_args
    avail_scoring_fn_names += additional_scoring_fn_names

    # add the default scoring functions if not already in there
    for scoring_name in categorical_enable["__any__"]:
        if scoring_options[scoring_name] not in avail_scoring_fns:
            avail_scoring_fn_def_args.append(builtin_default_kwargs)
            avail_scoring_fns.append(scoring_options[scoring_name])
            s_label = scoring_name
            if s_label in avail_scoring_fn_names:
                s_label = "default_" + s_label
            avail_scoring_fn_names.append(s_label)
            if len(default_scoring_fns) == 0:
                default_scoring_fns.append(s_label)

    return avail_scoring_fns, avail_scoring_fn_def_args, avail_scoring_fn_names, default_scoring_fns


def get_scoring_fns(model, scoring_fns, scoring_kwargs=None):
    """
    Transform a list of scoring functions and names to a dictionary of scoring functions that are set up with kwargs
    defined in scoring_kwargs.
    
    Arguments
        model: Kipoi Model object
        scoring_fns: A list of scoring functions or strings of scoring functions. 
        scoring_kwargs: Either list of length 0 or a list of dicts (kwargs) with the same length of scoring_fns and the 
        same order. If an entry of scoring_fns is a string then the scoring function will be initialised with the 
        corresponding entry in scoring_kwargs. If an entry of scoring_fns is a callable the corresponding entry in 
        scoring_kwargs will be ignored.
    """
    # get the scoring methods according to the model definition
    avail_scoring_fns, avail_scoring_fn_def_args, avail_scoring_fn_names, \
    default_scoring_fns = get_avail_scoring_fns(model)

    errmsg_scoring_kwargs = "When using `scoring_kwargs` a kwargs dictionary for every entry in " \
                            "`scoring_kwargs` must be given."
    if scoring_kwargs is None:
        scoring_kwargs = []

    dts = {}
    if len(scoring_fns) >= 1:
        # Check if all scoring functions should be used:
        if scoring_fns == ["all"]:
            if len(scoring_kwargs) >= 1:
                raise ValueError("`scoring_kwargs` cannot be defined when `scoring_fns` == ['all']!")
            for arg_iter, k in enumerate(avail_scoring_fn_names):
                si = avail_scoring_fn_names.index(k)
                # get the default kwargs
                kwargs = avail_scoring_fn_def_args[si]
                if kwargs is None:
                    raise ValueError("No default kwargs for scoring function: %s"
                                     " `scoring_fns` == ['all'] cannot be used. "
                                     " Please select the desired scoring functions explicitely and also define "
                                     "`scoring_kwargs`." % (k))
                # instantiate the scoring fn
                dts[k] = avail_scoring_fns[si](**kwargs)
        else:
            # if -k set check that length matches with -s
            if len(scoring_kwargs) >= 1:
                if not len(scoring_fns) == len(scoring_kwargs):
                    raise ValueError(errmsg_scoring_kwargs)
            for arg_iter, k in enumerate(scoring_fns):
                if isinstance(k, six.string_types):
                    # if -s set check is available for model
                    if k in avail_scoring_fn_names:
                        si = avail_scoring_fn_names.index(k)
                        # get the default kwargs
                        kwargs = avail_scoring_fn_def_args[si]
                        # if the user has set scoring function kwargs then load them here.
                        if len(scoring_kwargs) >= 1:
                            # all the {}s in -k replace by their defaults, if the default is None
                            # raise exception with the corrsponding scoring function label etc.
                            if len(scoring_kwargs[arg_iter]) != 0:
                                kwargs = scoring_kwargs[arg_iter]
                        if kwargs is None:
                            raise ValueError("No kwargs were given for scoring function %s"
                                             " with no defaults but required argmuents. "
                                             "Please also define sel_scoring_kwargs." % (k))
                        # instantiate the scoring fn
                        dts[k] = avail_scoring_fns[si](**kwargs)
                    else:
                        logger.warn("Cannot choose scoring function %s. "
                                    "Model only supports: %s." % (k, str(avail_scoring_fn_names)))
                elif callable(k):
                    # TODO: Add tests
                    if not hasattr(k, "__name__"):
                        raise Exception("scoring functions that are passed as callables have to have a __name__ "
                                        "attribute.")
                    dts[k.__name__] = k

    # if -s not set use all defaults
    elif len(default_scoring_fns) != 0:
        for arg_iter, k in enumerate(default_scoring_fns):
            si = avail_scoring_fn_names.index(k)
            kwargs = avail_scoring_fn_def_args[si]
            dts[k] = avail_scoring_fns[si](**kwargs)

    if len(dts) == 0:
        raise Exception("No scoring method was chosen!")

    return dts



from kipoi.postprocessing.variant_effects.scores import Ref, Alt, Diff, LogitRef, LogitAlt, Logit, DeepSEA_effect
from .parsers import KipoiVCFParser
from .snv_predict import predict_snvs, analyse_model_preds, score_variants
from .utils import ModelInfoExtractor, SnvPosRestrictedRg, SnvCenteredRg, ensure_tabixed_vcf, VcfWriter, \
    BedOverlappingRg
from .mutation_map import MutationMap
# from .mutation_map import _generate_mutation_map, MutationMapDrawer
